"""Scraper for http://www.football-data.co.uk/data.php."""
import itertools
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

import pandas as pd

from ._common import BaseRequestsReader, make_game_id
from ._config import DATA_DIR, NOCACHE, NOSTORE, TEAMNAME_REPLACEMENTS

MATCH_HISTORY_DATA_DIR = DATA_DIR / 'MatchHistory'
MATCH_HISTORY_API = 'https://www.football-data.co.uk'


class MatchHistory(BaseRequestsReader):
    """Provides pd.DataFrames from CSV files available at http://www.football-data.co.uk/data.php.

    Data will be downloaded as necessary and cached locally in
    ``~/soccerdata/data/MatchHistory``.

    Parameters
    ----------
    leagues : string or iterable
        IDs of leagues to include.
    seasons : string, int or list
        Seasons to include. Supports multiple formats.
        Examples: '16-17'; 2016; '2016-17'; [14, 15, 16]
    proxy : 'tor' or dict or list(dict) or callable, optional
        Use a proxy to hide your IP address. Valid options are:
            - "tor": Uses the Tor network. Tor should be running in
              the background on port 9050.
            - dict: A dictionary with the proxy to use. The dict should be
              a mapping of supported protocols to proxy addresses. For example::

                  {
                      'http': 'http://10.10.1.10:3128',
                      'https': 'http://10.10.1.10:1080',
                  }

            - list(dict): A list of proxies to choose from. A different proxy will
              be selected from this list after failed requests, allowing rotating
              proxies.
            - callable: A function that returns a valid proxy. This function will
              be called after failed requests, allowing rotating proxies.
    no_cache : bool
        If True, will not use cached data.
    no_store : bool
        If True, will not store downloaded data.
    data_dir : Path, optional
        Path to directory where data will be cached.
    """

    def __init__(
        self,
        leagues: Optional[Union[str, List[str]]] = None,
        seasons: Optional[Union[str, int, List]] = None,
        proxy: Optional[
            Union[str, Dict[str, str], List[Dict[str, str]], Callable[[], Dict[str, str]]]
        ] = None,
        no_cache: bool = NOCACHE,
        no_store: bool = NOSTORE,
        data_dir: Path = MATCH_HISTORY_DATA_DIR,
    ):
        super().__init__(
            leagues=leagues, proxy=proxy, no_cache=no_cache, no_store=no_store, data_dir=data_dir
        )
        self.seasons = seasons  # type: ignore

    def read_games(self) -> pd.DataFrame:
        """Retrieve game history for the selected leagues and seasons.

        Column names are explained here: http://www.football-data.co.uk/notes.txt

        Returns
        -------
        pd.DataFrame
        """
        urlmask = MATCH_HISTORY_API + '/mmz4281/{}/{}.csv'
        filemask = '{}_{}.csv'
        col_rename = {
            'Div': 'league',
            'Date': 'date',
            'Time': 'time',
            'HomeTeam': 'home_team',
            'AwayTeam': 'away_team',
            'Referee': 'referee',
        }

        df_list = []
        for lkey, skey in itertools.product(self._selected_leagues.values(), self.seasons):
            filepath = self.data_dir / filemask.format(lkey, skey)
            url = urlmask.format(skey, lkey)
            current_season = not self._is_complete(lkey, skey)
            reader = self.get(url, filepath, no_cache=current_season)

            df_list.append(
                pd.read_csv(
                    reader,
                    encoding='ISO-8859-1',
                ).assign(season=skey)
            )

        df = (
            pd.concat(df_list, sort=False)
            .rename(columns=col_rename)
            .assign(date=lambda x: pd.to_datetime(x["date"] + ' ' + x['time']))
            .drop("time", axis=1)
            .pipe(self._translate_league)
            .replace(
                {
                    'home_team': TEAMNAME_REPLACEMENTS,
                    'away_team': TEAMNAME_REPLACEMENTS,
                }
            )
            .dropna(subset=['home_team', 'away_team'])
        )

        df['game'] = df.apply(make_game_id, axis=1)
        df.set_index(['league', 'season', 'game'], inplace=True)
        df.sort_index(inplace=True)
        return df



"""Scraper for http://sofifa.com."""
import re
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

import pandas as pd
from lxml import html

from ._common import BaseRequestsReader, standardize_colnames
from ._config import DATA_DIR, NOCACHE, NOSTORE, TEAMNAME_REPLACEMENTS

SO_FIFA_DATADIR = DATA_DIR / "SoFIFA"
SO_FIFA_API = "https://sofifa.com"


class SoFIFA(BaseRequestsReader):
    """Provides pd.DataFrames from data at http://sofifa.com.

    Data will be downloaded as necessary and cached locally in
    ``~/soccerdata/data/SoFIFA``.

    Parameters
    ----------
    leagues : string or iterable, optional
        IDs of leagues to include.
    seasons : string, int or list, optional
        Seasons to include. Supports multiple formats.
        Examples: '16-17'; 2016; '2016-17'; [14, 15, 16]
    proxy : 'tor' or dict or list(dict) or callable, optional
        Use a proxy to hide your IP address. Valid options are:
            - "tor": Uses the Tor network. Tor should be running in
              the background on port 9050.
            - dict: A dictionary with the proxy to use. The dict should be
              a mapping of supported protocols to proxy addresses. For example::

                  {
                      'http': 'http://10.10.1.10:3128',
                      'https': 'http://10.10.1.10:1080',
                  }

            - list(dict): A list of proxies to choose from. A different proxy will
              be selected from this list after failed requests, allowing rotating
              proxies.
            - callable: A function that returns a valid proxy. This function will
              be called after failed requests, allowing rotating proxies.
    no_cache : bool
        If True, will not use cached data.
    no_store : bool
        If True, will not store downloaded data.
    data_dir : Path
        Path to directory where data will be cached.
    """

    def __init__(
        self,
        leagues: Optional[Union[str, List[str]]] = None,
        seasons: Optional[Union[str, int, List]] = None,
        proxy: Optional[
            Union[str, Dict[str, str], List[Dict[str, str]], Callable[[], Dict[str, str]]]
        ] = None,
        no_cache: bool = NOCACHE,
        no_store: bool = NOSTORE,
        data_dir: Path = SO_FIFA_DATADIR,
    ):
        """Initialize SoFIFA reader."""
        super().__init__(
            leagues=leagues,
            proxy=proxy,
            no_cache=no_cache,
            no_store=no_store,
            data_dir=data_dir,
        )
        self.rate_limit = 2
        self.seasons = seasons  # type: ignore

    def read_leagues(self) -> pd.DataFrame:
        """Retrieve selected leagues from the datasource.

        Returns
        -------
        pd.DataFrame
        """
        # read html page (overview)
        filepath = self.data_dir / "leagues.html"
        reader = self.get(SO_FIFA_API, filepath)

        # extract league links
        leagues = []
        tree = html.parse(reader)
        for node in tree.xpath("//select[@id='choices-lg']/optgroup/option"):
            leagues.append(
                {
                    "league_id": int(node.get("value")),
                    "league": node.text,
                }
            )
        df = pd.DataFrame(leagues).pipe(self._translate_league).set_index("league").sort_index()
        return df[df.index.isin(self._selected_leagues.keys())]

    def read_teams(self) -> pd.DataFrame:
        """Retrieve teams from the datasource for the selected leagues.

        Returns
        -------
        pd.DataFrame
        """
        # build url
        urlmask = SO_FIFA_API + "/teams?lg={}&v={}"
        filemask = "teams_{}_{}.html"

        # get league IDs
        leagues = self.read_leagues()

        # collect teams
        teams = []
        for lkey, _ in self._selected_leagues.items():
            league_id = leagues.at[lkey, "league_id"]
            for skey in self.seasons:
                # read html page (league overview)
                season_id = skey[:2]
                filepath = self.data_dir / filemask.format(lkey, skey)
                url = urlmask.format(league_id, season_id)
                reader = self.get(url, filepath)

                # extract team links
                tree = html.parse(reader)
                pat_team = re.compile(r"\/team\/(\d+)\/[\w-]+\/")
                for node in tree.xpath("//a[contains(@href,'/team/')]"):
                    # extract team IDs from links
                    teams.append(
                        {
                            "team_id": int(
                                re.search(pat_team, node.get("href")).group(1)  # type: ignore
                            ),
                            "team": node.xpath(".//div")[0].text,
                            "league": lkey,
                            "season": skey,
                        }
                    )

        # return data frame
        df = (
            pd.DataFrame(teams)
            .replace({"team": TEAMNAME_REPLACEMENTS})
            .set_index(["league", "season", "team"])
            .sort_index()
        )
        return df

    def read_players(self) -> pd.DataFrame:
        """Retrieve players from the datasource for the selected leagues.

        Returns
        -------
        pd.DataFrame
        """
        # build url
        urlmask = SO_FIFA_API + "/team/{}?v={}"
        filemask = str(self.data_dir / "players_{}_{}.html")

        # get team IDs
        teams = self.read_teams().reset_index()

        # collect players
        players = []
        for _, team in teams.iterrows():
            season_id = team.season[:2]
            team_name = team.team
            # read html page (team overview)
            filepath = self.data_dir / filemask.format(team_name, season_id)
            url = urlmask.format(team["team_id"], season_id)
            reader = self.get(url, filepath)

            # extract player links
            tree = html.parse(reader)
            pat_player = re.compile(r"\/player\/(\d+)\/[\w-]+\/")
            for node in tree.xpath("//a[contains(@href,'/player/') and @title]"):
                # extract player IDs from links
                # extract player names from links
                players.append(
                    {
                        "player_id": int(
                            re.search(pat_player, node.get("href")).group(1)  # type: ignore
                        ),
                        "player": node.get("title"),
                        "team": team_name,
                        "league": team.league,
                        "season": team.season,
                    }
                )

        # return data frame
        df = pd.DataFrame(players).set_index(["league", "season", "team", "player"]).sort_index()
        return df

    def read_ratings(self) -> pd.DataFrame:
        """Retrieve ratings from the datasource for the selected leagues.

        Returns
        -------
        pd.DataFrame
        """
        # build url
        urlmask = SO_FIFA_API + "/player/{}?v={}"
        filemask = "player_{}_{}.html"

        # get player IDs
        players = self.read_players().reset_index()

        # prepare empty data frame
        ratings = []

        # define labels to use for score extraction from player profile pages
        score_labels = [
            "Overall Rating",
            "Potential",
            "Crossing",
            "Finishing",
            "Heading Accuracy",
            "Short Passing",
            "Volleys",
            "Dribbling",
            "Curve",
            "FK Accuracy",
            "Long Passing",
            "Ball Control",
            "Acceleration",
            "Sprint Speed",
            "Agility",
            "Reactions",
            "Balance",
            "Shot Power",
            "Jumping",
            "Stamina",
            "Strength",
            "Long Shots",
            "Aggression",
            "Interceptions",
            "Positioning",
            "Vision",
            "Penalties",
            "Composure",
            "Marking",
            "Standing Tackle",
            "Sliding Tackle",
            "GK Diving",
            "GK Handling",
            "GK Kicking",
            "GK Positioning",
            "GK Reflexes",
        ]

        for _, player in players.iterrows():
            # read html page (player overview)
            player_name = player.player
            filepath = self.data_dir / filemask.format(player_name, player.season)
            url = urlmask.format(player["player_id"], player.season[:2])
            reader = self.get(url, filepath)

            # extract scores one-by-one
            tree = html.parse(reader)
            scores = {
                "player": player_name,
                "league": player.league,
                "season": player.season,
            }
            for s in score_labels:
                nodes = tree.xpath(
                    "(//li[not(self::script)] | //div)"
                    f"[.//text()[contains(.,'{s}')]]"
                    "/span[contains(@class, 'tag')]"
                )
                # for multiple matches, only accept first match
                if len(nodes) >= 1:
                    scores[s] = nodes[0].text.strip()
                # if there's no match, put NA
                else:
                    scores[s] = None
            ratings.append(scores)
        # return data frame
        df = (
            pd.DataFrame(ratings)
            .pipe(standardize_colnames)
            .set_index(["league", "season", "player"])
            .sort_index()
        )
        return df



import io
import json
import pprint
import random
import re
import time
import warnings
from abc import ABC, abstractmethod
from datetime import date, datetime, timedelta
from pathlib import Path
from typing import IO, Callable, Dict, Iterable, List, Optional, Union

import numpy as np
import pandas as pd
import requests
import undetected_chromedriver as uc
from dateutil.relativedelta import relativedelta
from selenium.common.exceptions import WebDriverException

from ._config import DATA_DIR, LEAGUE_DICT, logger


class BaseReader(ABC):
    """Base class for data readers.

    Parameters
    ----------
    leagues : str or list of str, optional
        The leagues to read. If None, all available leagues are read.
    proxy : 'tor' or or dict or list(dict) or callable, optional
        Use a proxy to hide your IP address. Valid options are:
            - "tor": Uses the Tor network. Tor should be running in
              the background on port 9050.
            - dict: A dictionary with the proxy to use. The dict should be
              a mapping of supported protocols to proxy addresses. For example::

                  {
                      'http': 'http://10.10.1.10:3128',
                      'https': 'http://10.10.1.10:1080',
                  }

            - list(dict): A list of proxies to choose from. A different proxy will
              be selected from this list after failed requests, allowing rotating
              proxies.
            - callable: A function that returns a valid proxy. This function will
              be called after failed requests, allowing rotating proxies.
    no_cache : bool
        If True, will not use cached data.
    no_store : bool
        If True, will not store downloaded data.
    data_dir : Path
        Path to directory where data will be cached.
    """

    def __init__(
        self,
        leagues: Optional[Union[str, List[str]]] = None,
        proxy: Optional[
            Union[str, Dict[str, str], List[Dict[str, str]], Callable[[], Dict[str, str]]]
        ] = None,
        no_cache: bool = False,
        no_store: bool = False,
        data_dir: Path = DATA_DIR,
    ):
        """Create a new data reader."""
        if isinstance(proxy, str) and proxy.lower() == "tor":
            self.proxy = lambda: {
                "http": "socks5://127.0.0.1:9050",
                "https": "socks5://127.0.0.1:9050",
            }
        elif isinstance(proxy, dict):
            self.proxy = lambda: proxy  # type: ignore
        elif isinstance(proxy, list):
            self.proxy = lambda: random.choice(proxy)  # type: ignore
        elif callable(proxy):
            self.proxy = proxy
        else:
            self.proxy = lambda: {}

        self._selected_leagues = leagues  # type: ignore
        self.no_cache = no_cache
        self.no_store = no_store
        self.data_dir = data_dir
        self.rate_limit = 0
        self.max_delay = 0
        if self.no_store:
            logger.info("Caching is disabled")
        else:
            logger.info("Saving cached data to %s", self.data_dir)
            self.data_dir.mkdir(parents=True, exist_ok=True)

    def get(
        self,
        url: str,
        filepath: Optional[Path] = None,
        max_age: Optional[Union[int, timedelta]] = None,
        no_cache: bool = False,
        var: Optional[str] = None,
    ) -> IO[bytes]:
        """Load data from `url`.

        By default, the source of `url` is downloaded and saved to `filepath`.
        If `filepath` exists, the `url` is not visited and the cached data is
        returned.

        Parameters
        ----------
        url : str
            URL to download.
        filepath : Path, optional
            Path to save downloaded file. If None, downloaded data is not cached.
        max_age : int for age in days, or timedelta object
            The max. age of locally cached file before re-download.
        no_cache : bool
            If True, will not use cached data. Overrides the class property.
        var : str, optional
            Return a javascript variable instead of the page source.

        Raises
        ------
        TypeError
            If max_age is not an integer or timedelta object.

        Returns
        -------
        io.BufferedIOBase
            File-like object of downloaded data.
        """
        is_cached = self._is_cached(filepath, max_age)
        if no_cache or self.no_cache or not is_cached:
            logger.debug("Scraping %s", url)
            return self._download_and_save(url, filepath, var)
        logger.debug("Retrieving %s from cache", url)
        assert filepath is not None
        return filepath.open(mode="rb")

    def _is_cached(
        self,
        filepath: Optional[Path] = None,
        max_age: Optional[Union[int, timedelta]] = None,
    ) -> bool:
        """Check if `filepath` contains valid cached data.

        Parameters
        ----------
        filepath : Path, optional
            Path where file should be cached. If None, return False.
        max_age : int for age in days, or timedelta object
            The max. age of locally cached file.

        Raises
        ------
        TypeError
            If max_age is not an integer or timedelta object.

        Returns
        -------
        bool
            True in case of a cache hit, otherwise False.
        """
        # Validate inputs
        if max_age is not None:
            if isinstance(max_age, int):
                _max_age = timedelta(days=max_age)
            elif isinstance(max_age, timedelta):
                _max_age = max_age
            else:
                raise TypeError("max_age must be of type int or datetime.timedelta")
        else:
            _max_age = None

        cache_invalid = False
        # Check if cached file is too old
        if _max_age is not None and filepath is not None and filepath.exists():
            last_modified = datetime.fromtimestamp(filepath.stat().st_mtime)
            now = datetime.now()
            if (now - last_modified) > _max_age:
                cache_invalid = True

        return not cache_invalid and filepath is not None and filepath.exists()

    @abstractmethod
    def _download_and_save(
        self,
        url: str,
        filepath: Optional[Path] = None,
        var: Optional[str] = None,
    ) -> IO[bytes]:
        """Download data at `url` to `filepath`.

        Parameters
        ----------
        url : str
            URL to download.
        filepath : Path, optional
            Path to save downloaded file. If None, downloaded data is not cached.
        var : str, optional
            Return a javascript variable instead of the page source.

        Returns
        -------
        io.BufferedIOBase
            File-like object of downloaded data.
        """

    @classmethod
    def available_leagues(cls) -> List[str]:
        """Return a list of league IDs available for this source."""
        return sorted(cls._all_leagues().keys())

    @classmethod
    def _all_leagues(cls) -> Dict[str, str]:
        """Return a dict mapping all canonical league IDs to source league IDs."""
        if not hasattr(cls, "_all_leagues_dict"):
            cls._all_leagues_dict = {  # type: ignore
                k: v[cls.__name__] for k, v in LEAGUE_DICT.items() if cls.__name__ in v
            }
        return cls._all_leagues_dict  # type: ignore

    @classmethod
    def _translate_league(cls, df: pd.DataFrame, col: str = "league") -> pd.DataFrame:
        """Map source league ID to canonical ID."""
        flip = {v: k for k, v in cls._all_leagues().items()}
        mask = ~df[col].isin(flip)
        df.loc[mask, col] = np.nan
        df[col] = df[col].replace(flip)
        return df

    @property
    def _selected_leagues(self) -> Dict[str, str]:
        """Return a dict mapping selected canonical league IDs to source league IDs."""
        return self._leagues_dict

    @_selected_leagues.setter
    def _selected_leagues(self, ids: Optional[Union[str, List[str]]] = None) -> None:
        if ids is None:
            self._leagues_dict = self._all_leagues()
        else:
            if len(ids) == 0:
                raise ValueError("Empty iterable not allowed for 'leagues'")
            if isinstance(ids, str):
                ids = [ids]
            tmp_league_dict = {}
            for i in ids:
                if i not in self._all_leagues():
                    raise ValueError(
                        f"""
                        Invalid league '{i}'. Valid leagues are:
                        { pprint.pformat(self.available_leagues()) }
                        """
                    )
                tmp_league_dict[i] = self._all_leagues()[i]
            self._leagues_dict = tmp_league_dict

    def _is_complete(self, league: str, season: str) -> bool:
        """Check if a season is complete."""
        if league in LEAGUE_DICT:
            league_dict = LEAGUE_DICT[league]
        else:
            flip = {v: k for k, v in self._all_leagues().items()}
            if league in flip:
                league_dict = LEAGUE_DICT[flip[league]]
            else:
                raise ValueError(f"Invalid league '{league}'")
        if "season_end" not in league_dict:
            season_ends = date(datetime.strptime(season[-2:], "%y").year, 7, 1)
        else:
            season_ends = date(
                datetime.strptime(season[-2:], "%y").year,
                datetime.strptime(league_dict["season_end"], "%b").month,
                1,
            ) + relativedelta(months=1)
        return date.today() >= season_ends

    @property
    def leagues(self) -> List[str]:
        """Return a list of selected leagues."""
        return list(self._leagues_dict.keys())

    @property
    def seasons(self) -> List[str]:
        """Return a list of selected seasons."""
        return self._season_ids

    @seasons.setter
    def seasons(self, seasons: Optional[Union[str, int, Iterable[Union[str, int]]]]) -> None:
        if seasons is None:
            logger.info("No seasons provided. Will retrieve data for the last 5 seasons.")
            year = datetime.today().year
            seasons = [f"{y-1}-{y}" for y in range(year, year - 6, -1)]
        if isinstance(seasons, str) or isinstance(seasons, int):
            seasons = [seasons]
        self._season_ids = [season_code(s) for s in seasons]


class BaseRequestsReader(BaseReader):
    """Base class for readers that use the Python requests module."""

    def __init__(
        self,
        leagues: Optional[Union[str, List[str]]] = None,
        proxy: Optional[
            Union[str, Dict[str, str], List[Dict[str, str]], Callable[[], Dict[str, str]]]
        ] = None,
        no_cache: bool = False,
        no_store: bool = False,
        data_dir: Path = DATA_DIR,
    ):
        """Initialize the reader."""
        super().__init__(
            no_cache=no_cache,
            no_store=no_store,
            leagues=leagues,
            proxy=proxy,
            data_dir=data_dir,
        )

        self._session = self._init_session()

    def _init_session(self) -> requests.Session:
        session = requests.Session()
        session.proxies.update(self.proxy())
        return session

    def _download_and_save(
        self,
        url: str,
        filepath: Optional[Path] = None,
        var: Optional[str] = None,
    ) -> IO[bytes]:
        """Download file at url to filepath. Overwrites if filepath exists."""
        for i in range(5):
            try:
                response = self._session.get(url, stream=True)
                time.sleep(self.rate_limit + random.random() * self.max_delay)
                response.raise_for_status()
                if not self.no_store and filepath is not None:
                    with filepath.open(mode="wb") as fh:
                        fh.write(response.content)
                return io.BytesIO(response.content)
            except Exception:
                logger.exception(
                    "Error while scraping %s. Retrying... (attempt %d of 5).", url, i + 1
                )
                self._session = self._init_session()
                continue

        raise ConnectionError("Could not download %s." % url)


class BaseSeleniumReader(BaseReader):
    """Base class for readers that use Selenium."""

    def __init__(
        self,
        leagues: Optional[Union[str, List[str]]] = None,
        proxy: Optional[
            Union[str, Dict[str, str], List[Dict[str, str]], Callable[[], Dict[str, str]]]
        ] = None,
        no_cache: bool = False,
        no_store: bool = False,
        data_dir: Path = DATA_DIR,
        path_to_browser: Optional[Path] = None,
        headless: bool = True,
    ):
        """Initialize the reader."""
        super().__init__(
            no_cache=no_cache,
            no_store=no_store,
            leagues=leagues,
            proxy=proxy,
            data_dir=data_dir,
        )
        self.path_to_browser = path_to_browser
        self.headless = headless

        try:
            self._driver = self._init_webdriver()
        except WebDriverException as e:
            logger.error(
                """
                The ChromeDriver was unable to initiate/spawn a new
                WebBrowser. You will not be able to scrape new data.
                %s
                """,
                e,
            )

    def _init_webdriver(self) -> "uc.Chrome":
        """Start the Selenium driver."""
        # Quit existing driver
        if hasattr(self, "_driver"):
            self._driver.quit()
        # Start a new driver
        chrome_options = uc.ChromeOptions()
        if self.headless:
            chrome_options.add_argument("--headless")
        if self.path_to_browser is not None:
            chrome_options.add_argument("--binary-location=" + str(self.path_to_browser))
        proxy = self.proxy()
        if len(proxy):
            proxy_str = ";".join(f"{prot}={url}" for prot, url in proxy.items())
            resolver_rules = "MAP * ~NOTFOUND , EXCLUDE 127.0.0.1"
            chrome_options.add_argument("--proxy-server=" + proxy_str)
            chrome_options.add_argument("--host-resolver-rules=" + resolver_rules)
        return uc.Chrome(options=chrome_options)

    def _download_and_save(  # noqa: C901
        self,
        url: str,
        filepath: Optional[Path] = None,
        var: Optional[str] = None,
    ) -> IO[bytes]:
        """Download file at url to filepath. Overwrites if filepath exists."""
        for i in range(5):
            try:
                self._driver.get(url)
                time.sleep(self.rate_limit + random.random() * self.max_delay)
                if "Incapsula incident ID" in self._driver.page_source:
                    raise WebDriverException(
                        "Your IP is blocked. Use tor or a proxy to continue scraping."
                    )
                if var is None:
                    response = self._driver.execute_script(
                        "return document.body.innerHTML;"
                    ).encode("utf-8")
                else:
                    response = json.dumps(self._driver.execute_script("return " + var)).encode(
                        "utf-8"
                    )
                if not self.no_store and filepath is not None:
                    filepath.parent.mkdir(parents=True, exist_ok=True)
                    with filepath.open(mode="wb") as fh:
                        fh.write(response)
                return io.BytesIO(response)
            except Exception:
                logger.exception(
                    "Error while scraping %s. Retrying... (attempt %d of 5).", url, i + 1
                )
                self._driver = self._init_webdriver()
                continue

        raise ConnectionError("Could not download %s." % url)


def season_code(season: Union[str, int]) -> str:  # noqa: C901
    """Convert a string or int to a season code like '1718'."""
    season = str(season)
    pat1 = re.compile(r"^[0-9]{4}$")  # 1994 | 9495
    pat2 = re.compile(r"^[0-9]{2}$")  # 94
    pat3 = re.compile(r"^[0-9]{4}-[0-9]{4}$")  # 1994-1995
    pat4 = re.compile(r"^[0-9]{4}/[0-9]{4}$")  # 1994/1995
    pat5 = re.compile(r"^[0-9]{4}-[0-9]{2}$")  # 1994-95
    pat6 = re.compile(r"^[0-9]{2}-[0-9]{2}$")  # 94-95

    if re.match(pat1, season):
        if int(season[2:]) == int(season[:2]) + 1:
            if season == "2021":
                msg = 'Season id "{}" is ambiguous: interpreting as "{}-{}"'.format(
                    season, season[:2], season[-2:]
                )
                warnings.warn(msg)
            return season  # 9495
        elif season[2:] == "99":
            return "".join([season[2:], "00"])  # 1999
        else:
            return "".join([season[-2:], f"{int(season[-2:]) + 1:02d}"])  # 1994
    elif re.match(pat2, season):
        if season == "99":
            return "".join([season, "00"])  # 99
        else:
            return "".join([season, f"{int(season) + 1:02d}"])  # 94
    elif re.match(pat3, season):
        return "".join([season[2:4], season[-2:]])  # 1994-1995
    elif re.match(pat4, season):
        return "".join([season[2:4], season[-2:]])  # 1994/1995
    elif re.match(pat5, season):
        return "".join([season[2:4], season[-2:]])  # 1994-95
    elif re.match(pat6, season):
        return "".join([season[:2], season[-2:]])  # 94-95
    else:
        return season


def make_game_id(row: pd.Series) -> str:
    """Return a game id based on date, home and away team."""
    if pd.isnull(row["date"]):
        game_id = "{}-{}".format(
            row["home_team"],
            row["away_team"],
        )
    else:
        game_id = "{} {}-{}".format(
            row["date"].strftime("%Y-%m-%d"),
            row["home_team"],
            row["away_team"],
        )
    return game_id


def standardize_colnames(df: pd.DataFrame, cols: Optional[List[str]] = None) -> pd.DataFrame:
    """Convert DataFrame column names to snake case."""

    def to_snake(name: str) -> str:
        name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
        name = re.sub("__([A-Z])", r"_\1", name)
        name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", name)
        return name.lower().replace("-", "_").replace(" ", "")

    if cols is None:
        cols = list(df.columns)

    return df.rename(columns={c: to_snake(c) for c in cols})


def get_proxy() -> Dict[str, str]:
    """Return a public proxy."""
    # list of free proxy apis
    # protocols: http, https, socks4 and socks5
    list_of_proxy_content = [
        "https://proxylist.geonode.com/api/proxy-list?sort_by=lastChecked&sort_type=desc",
    ]

    # extracting json data from this list of proxies
    full_proxy_list = []
    for proxy_url in list_of_proxy_content:
        proxy_json = json.loads(requests.get(proxy_url).text)["data"]
        full_proxy_list.extend(proxy_json)

        if not full_proxy_list:
            logger.info("There are currently no proxies available. Exiting...")
            return {}
        else:
            logger.info(f"Found {len(full_proxy_list)} proxy servers. Checking...\n")

    # creating proxy dict
    final_proxy_list = []
    for proxy in full_proxy_list:
        protocol = proxy["protocols"][0]
        ip_ = proxy["ip"]
        port = proxy["port"]

        proxy = {
            "https": protocol + "://" + ip_ + ":" + port,
            "http": protocol + "://" + ip_ + ":" + port,
        }

        final_proxy_list.append(proxy)

    # trying proxy
    for proxy in final_proxy_list:
        if check_proxy(proxy):
            return proxy

    logger.info("There are currently no proxies available. Exiting...")
    return {}


def check_proxy(proxy: dict) -> bool:
    """Check if proxy is working."""
    try:
        r0 = requests.get("https://ipinfo.io/json", proxies=proxy, timeout=15)
        return r0.status_code == 200
    except Exception as error:
        logger.error(f"BAD PROXY: Reason: {str(error)}\n")
        return False



"""Scraper for api.clubelo.com."""
import re
from datetime import datetime, timedelta
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

import pandas as pd
from unidecode import unidecode

from ._common import BaseRequestsReader, standardize_colnames
from ._config import DATA_DIR, NOCACHE, NOSTORE, TEAMNAME_REPLACEMENTS

CLUB_ELO_DATADIR = DATA_DIR / "ClubElo"
CLUB_ELO_API = "http://api.clubelo.com"


class ClubElo(BaseRequestsReader):
    """Provides pd.DataFrames from CSV API at http://api.clubelo.com.

    Data will be downloaded as necessary and cached locally in
    ``~/soccerdata/data/ClubElo``.

    Since the source does not provide league names, this class will not filter
    by league. League names will be inserted from the other sources where
    available. Leagues that are only covered by clubelo.com will have NaN
    values.

    Parameters
    ----------
    proxy : 'tor' or dict or list(dict) or callable, optional
        Use a proxy to hide your IP address. Valid options are:
            - "tor": Uses the Tor network. Tor should be running in
              the background on port 9050.
            - dict: A dictionary with the proxy to use. The dict should be
              a mapping of supported protocols to proxy addresses. For example::

                  {
                      'http': 'http://10.10.1.10:3128',
                      'https': 'http://10.10.1.10:1080',
                  }

            - list(dict): A list of proxies to choose from. A different proxy will
              be selected from this list after failed requests, allowing rotating
              proxies.
            - callable: A function that returns a valid proxy. This function will
              be called after failed requests, allowing rotating proxies.
    no_cache : bool
        If True, will not use cached data.
    no_store : bool
        If True, will not store downloaded data.
    data_dir : Path
        Path to directory where data will be cached.
    """

    def __init__(
        self,
        proxy: Optional[
            Union[str, Dict[str, str], List[Dict[str, str]], Callable[[], Dict[str, str]]]
        ] = None,
        no_cache: bool = NOCACHE,
        no_store: bool = NOSTORE,
        data_dir: Path = CLUB_ELO_DATADIR,
    ):
        """Initialize a new ClubElo reader."""
        super().__init__(no_cache=no_cache, no_store=no_store, data_dir=data_dir)

    def read_by_date(self, date: Optional[Union[str, datetime]] = None) -> pd.DataFrame:
        """Retrieve ELO scores for all teams at specified date.

        Elo scores are available as early as 1939. Values before 1960 should
        be considered provisional.

        Parameters
        ----------
        date : datetime object or string like 'YYYY-MM-DD'
            Date for which to retrieve ELO scores. If no date is specified,
            get today's scores.

        Returns
        -------
        pd.DataFrame
        """
        if not date:
            date = datetime.today()
        elif isinstance(date, str):
            date = datetime.strptime(date, "%Y-%m-%d")
        else:
            pass  # Assume datetime object

        datestring = date.strftime("%Y-%m-%d")
        filepath = self.data_dir / f"{datestring}.csv"
        url = f"{CLUB_ELO_API}/{datestring}"

        data = self.get(url, filepath)

        df = (
            pd.read_csv(
                data, parse_dates=["From", "To"], infer_datetime_format=True, dayfirst=False
            )
            .pipe(standardize_colnames)
            .rename(columns={"club": "team"})
            .replace({"team": TEAMNAME_REPLACEMENTS})
            .replace("None", float("nan"))
            .assign(rank=lambda x: x["rank"].astype("float"))
            .assign(league=lambda x: x["country"] + "_" + x["level"].astype(str))
            .pipe(self._translate_league)
            .reset_index(drop=True)
            .set_index("team")
        )
        return df

    def read_team_history(
        self, team: str, max_age: Union[int, timedelta] = 1
    ) -> Optional[pd.DataFrame]:
        """Retrieve full ELO history for one club.

        For the exact spelling of a club's name, check the result
        of :func:`~soccerdata.ClubElo.read_by_date` or
        `clubelo.com <http://clubelo.com/Ranking>`__. You can also use
        alternative team names specified in `teamname_replacements.json`.
        Values before 1960 should be considered provisional.

        Parameters
        ----------
        team : str
            The club's name
        max_age : int for age in days, or timedelta object
            The max. age of locally cached file before re-download.

        Raises
        ------
        TypeError
            If max_age is not an integer or timedelta object.
        ValueError
            If no ratings for the given team are available.

        Returns
        -------
        pd.DataFrame
        """
        teams_to_check = [k for k, v in TEAMNAME_REPLACEMENTS.items() if v == team]
        teams_to_check.append(team)

        for i, _ in enumerate(teams_to_check):
            teams_to_check[i] = unidecode(teams_to_check[i])
            teams_to_check[i] = re.sub(r"[\s']", "", teams_to_check[i])

        for _team in teams_to_check:
            filepath = self.data_dir / f"{_team}.csv"
            url = f"{CLUB_ELO_API}/{_team}"
            data = self.get(url, filepath, max_age)

            df = (
                pd.read_csv(
                    data,
                    parse_dates=["From", "To"],
                    infer_datetime_format=True,
                    dayfirst=False,
                )
                .pipe(standardize_colnames)
                .rename(columns={"club": "team"})
                .replace("None", float("nan"))
                .assign(rank=lambda x: x["rank"].astype("float"))
                .set_index("from")
                .sort_index()
            )

            if len(df) > 0:
                # clubelo.com returns a CSV with just a header for nonexistent club
                df.replace({"team": TEAMNAME_REPLACEMENTS}, inplace=True)
                return df

        raise ValueError(f"No data found for team {team}")



"""Scraper for http://fbref.com."""
import itertools
import warnings
from functools import reduce
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

import pandas as pd
from lxml import etree, html

from ._common import (
    BaseRequestsReader,
    make_game_id,
    season_code,
    standardize_colnames,
)
from ._config import DATA_DIR, NOCACHE, NOSTORE, TEAMNAME_REPLACEMENTS, logger

FBREF_DATADIR = DATA_DIR / "FBref"
FBREF_API = "https://fbref.com"

BIG_FIVE_DICT = {
    "Serie A": "ITA-Serie A",
    "Ligue 1": "FRA-Ligue 1",
    "La Liga": "ESP-La Liga",
    "Premier League": "ENG-Premier League",
    "Bundesliga": "GER-Bundesliga",
}


class FBref(BaseRequestsReader):
    """Provides pd.DataFrames from data at http://fbref.com.

    Data will be downloaded as necessary and cached locally in
    ``~/soccerdata/data/FBref``.

    Parameters
    ----------
    leagues : string or iterable, optional
        IDs of leagues to include. For efficiently reading data from the Top-5
        European leagues, use "Big 5 European Leagues Combined".
    seasons : string, int or list, optional
        Seasons to include. Supports multiple formats.
        Examples: '16-17'; 2016; '2016-17'; [14, 15, 16]
    proxy : 'tor' or dict or list(dict) or callable, optional
        Use a proxy to hide your IP address. Valid options are:
            - "tor": Uses the Tor network. Tor should be running in
              the background on port 9050.
            - dict: A dictionary with the proxy to use. The dict should be
              a mapping of supported protocols to proxy addresses. For example::

                  {
                      'http': 'http://10.10.1.10:3128',
                      'https': 'http://10.10.1.10:1080',
                  }

            - list(dict): A list of proxies to choose from. A different proxy will
              be selected from this list after failed requests, allowing rotating
              proxies.
            - callable: A function that returns a valid proxy. This function will
              be called after failed requests, allowing rotating proxies.
    no_cache : bool
        If True, will not use cached data.
    no_store : bool
        If True, will not store downloaded data.
    data_dir : Path
        Path to directory where data will be cached.
    """

    def __init__(
        self,
        leagues: Optional[Union[str, List[str]]] = None,
        seasons: Optional[Union[str, int, List]] = None,
        proxy: Optional[
            Union[str, Dict[str, str], List[Dict[str, str]], Callable[[], Dict[str, str]]]
        ] = None,
        no_cache: bool = NOCACHE,
        no_store: bool = NOSTORE,
        data_dir: Path = FBREF_DATADIR,
    ):
        """Initialize FBref reader."""
        super().__init__(
            leagues=leagues,
            proxy=proxy,
            no_cache=no_cache,
            no_store=no_store,
            data_dir=data_dir,
        )
        self.rate_limit = 3
        self.seasons = seasons  # type: ignore
        # check if all top 5 leagues are selected
        selected_leagues = set(self._leagues_dict.keys())
        if set(BIG_FIVE_DICT.values()).issubset(selected_leagues):
            warnings.warn(
                "You are trying to scrape data for all of the Big 5 European leagues. "
                "This can be done more efficiently by setting "
                "leagues='Big 5 European Leagues Combined'."
            )

    @property
    def leagues(self) -> List[str]:
        """Return a list of selected leagues."""
        selected_leagues = set(self._leagues_dict.keys())
        if "Big 5 European Leagues Combined" in selected_leagues:
            selected_leagues -= set(BIG_FIVE_DICT.values())
        return list(selected_leagues)

    @classmethod
    def _all_leagues(cls) -> Dict[str, str]:
        """Return a dict mapping all canonical league IDs to source league IDs."""
        res = super()._all_leagues()
        res.update({"Big 5 European Leagues Combined": "Big 5 European Leagues Combined"})
        return res

    def read_leagues(self) -> pd.DataFrame:
        """Retrieve selected leagues from the datasource.

        Returns
        -------
        pd.DataFrame
        """
        url = f"{FBREF_API}/en/comps/"
        filepath = self.data_dir / "leagues.html"
        reader = self.get(url, filepath)

        # extract league links
        leagues = []
        tree = html.parse(reader)
        for table in tree.xpath("//table[contains(@id, 'comps')]"):
            df_table = pd.read_html(etree.tostring(table, method="html"))[0]
            df_table["url"] = table.xpath(".//th[@data-stat='league_name']/a/@href")
            leagues.append(df_table)
        df = (
            pd.concat(leagues)
            .pipe(standardize_colnames)
            .rename(columns={"competition_name": "league"})
            .pipe(self._translate_league)
            .drop_duplicates(subset="league")
            .set_index("league")
            .sort_index()
        )
        df["first_season"] = df["first_season"].apply(season_code)
        df["last_season"] = df["last_season"].apply(season_code)
        df["country"] = df["country"].apply(
            lambda x: x.split(" ")[1] if isinstance(x, str) else None
        )
        return df[df.index.isin(self.leagues)]

    def read_seasons(self) -> pd.DataFrame:
        """Retrieve the selected seasons for the selected leagues.

        Returns
        -------
        pd.DataFrame
        """
        filemask = "seasons_{}.html"
        df_leagues = self.read_leagues()

        seasons = []
        for lkey, league in df_leagues.iterrows():
            url = FBREF_API + league.url
            filepath = self.data_dir / filemask.format(lkey)
            reader = self.get(url, filepath)

            # extract season links
            tree = html.parse(reader)
            df_table = pd.read_html(etree.tostring(tree), attrs={"id": "seasons"})[0]
            df_table["url"] = tree.xpath(
                "//table[@id='seasons']//th[@data-stat='year_id' or @data-stat='year']/a/@href"
            )
            # Override the competition name or add if missing
            df_table["Competition Name"] = lkey
            # Some tournaments have a "year" column instead of "season"
            if "Year" in df_table.columns:
                df_table.rename(columns={"Year": "Season"}, inplace=True)
            # Get the competition format
            if "Final" in df_table.columns:
                df_table["Format"] = "elimination"
            else:
                df_table["Format"] = "round-robin"
            seasons.append(df_table)

        df = pd.concat(seasons).pipe(standardize_colnames)
        df = df.rename(columns={"competition_name": "league"})
        df["season"] = df["season"].apply(season_code)
        # if both a 20xx and 19xx season are available, drop the 19xx season
        df.drop_duplicates(subset=["league", "season"], keep="first", inplace=True)
        df = df.set_index(["league", "season"]).sort_index()
        return df.loc[
            df.index.isin(itertools.product(self.leagues, self.seasons)), ["format", "url"]
        ]

    def read_team_season_stats(  # noqa: C901
        self, stat_type: str = "standard", opponent_stats: bool = False
    ) -> pd.DataFrame:
        """Retrieve teams from the datasource for the selected leagues.

        The following stat types are available:
            * 'standard'
            * 'keeper'
            * 'keeper_adv'
            * 'shooting'
            * 'passing'
            * 'passing_types'
            * 'goal_shot_creation'
            * 'defense'
            * 'possession'
            * 'playing_time'
            * 'misc'

        Parameters
        ----------
        stat_type: str
            Type of stats to retrieve.
        opponent_stats: bool
            If True, will retrieve opponent stats.

        Raises
        ------
        TypeError
            If ``stat_type`` is not valid.

        Returns
        -------
        pd.DataFrame
        """
        team_stats = [
            "standard",
            "keeper",
            "keeper_adv",
            "shooting",
            "passing",
            "passing_types",
            "goal_shot_creation",
            "defense",
            "possession",
            "playing_time",
            "misc",
        ]

        filemask = "teams_{}_{}_{}.html"

        if stat_type not in team_stats:
            raise TypeError(f"Invalid argument: stat_type should be in {team_stats}")

        if stat_type == "standard":
            page = "stats"
        elif stat_type == "keeper":
            page = "keepers"
        elif stat_type == "keeper_adv":
            page = "keepersadv"
        elif stat_type == "goal_shot_creation":
            page = "gca"
            stat_type = "gca"
        elif stat_type == "playing_time":
            page = "playingtime"
        else:
            page = stat_type

        if opponent_stats:
            stat_type += "_against"
        else:
            stat_type += "_for"

        # get league IDs
        seasons = self.read_seasons()

        # collect teams
        teams = []
        for (lkey, skey), season in seasons.iterrows():
            big_five = lkey == "Big 5 European Leagues Combined"
            tournament = season["format"] == "elimination"
            # read html page (league overview)
            filepath = self.data_dir / filemask.format(
                lkey, skey, stat_type if big_five else "all"
            )
            url = (
                FBREF_API
                + "/".join(season.url.split("/")[:-1])
                + (f"/{page}/squads/" if big_five else f"/{page}/" if tournament else "/")
                + season.url.split("/")[-1]
            )
            reader = self.get(url, filepath)

            # parse HTML and select table
            tree = html.parse(reader).xpath(
                f"//table[@id='stats_teams_{stat_type}' or @id='stats_squads_{stat_type}']"
            )[0]
            # remove icons
            for elem in tree.xpath("//span"):
                elem.getparent().remove(elem)
            # parse table
            df_table = pd.read_html(etree.tostring(tree))[0]
            df_table["league"] = lkey
            df_table["season"] = skey
            df_table["url"] = tree.xpath(".//*[@data-stat='team']/a/@href")
            if big_five:
                df_table["league"] = (
                    df_table.xs("Comp", axis=1, level=1).squeeze().map(BIG_FIVE_DICT)
                )
                df_table.drop("Comp", axis=1, level=1, inplace=True)
                df_table.drop("Rk", axis=1, level=1, inplace=True)
            teams.append(df_table)

        # return data frame
        df = (
            _concat(teams)
            .rename(columns={"Squad": "team"})
            .replace({"team": TEAMNAME_REPLACEMENTS})
            .set_index(["league", "season", "team"])
            .sort_index()
        )
        return df

    def read_player_season_stats(self, stat_type: str = "standard") -> pd.DataFrame:  # noqa: C901
        """Retrieve players from the datasource for the selected leagues.

        The following stat types are available:
            * 'standard'
            * 'shooting'
            * 'passing'
            * 'passing_types'
            * 'goal_shot_creation'
            * 'defense'
            * 'possession'
            * 'playing_time'
            * 'misc'
            * 'keeper'
            * 'keeper_adv'

        Parameters
        ----------
        stat_type :str
            Type of stats to retrieve.

        Raises
        ------
        TypeError
            If ``stat_type`` is not valid.

        Returns
        -------
        pd.DataFrame
        """
        player_stats = [
            "standard",
            "keeper",
            "keeper_adv",
            "shooting",
            "passing",
            "passing_types",
            "goal_shot_creation",
            "defense",
            "possession",
            "playing_time",
            "misc",
        ]

        filemask = "players_{}_{}_{}.html"

        if stat_type not in player_stats:
            raise TypeError(f"Invalid argument: stat_type should be in {player_stats}")

        if stat_type == "standard":
            page = "stats"
        elif stat_type == "goal_shot_creation":
            page = "gca"
            stat_type = "gca"
        elif stat_type == "playing_time":
            page = "playingtime"
        elif stat_type == "keeper":
            page = "keepers"
        elif stat_type == "keeper_adv":
            page = "keepersadv"
        else:
            page = stat_type

        # get league IDs
        seasons = self.read_seasons()

        # collect players
        players = []
        for (lkey, skey), season in seasons.iterrows():
            big_five = lkey == "Big 5 European Leagues Combined"
            filepath = self.data_dir / filemask.format(lkey, skey, stat_type)
            url = (
                FBREF_API
                + "/".join(season.url.split("/")[:-1])
                + f"/{page}"
                + ("/players/" if big_five else "/")
                + season.url.split("/")[-1]
            )
            reader = self.get(url, filepath)
            tree = html.parse(reader)
            # remove icons
            for elem in tree.xpath("//td[@data-stat='comp_level']//span"):
                elem.getparent().remove(elem)
            if big_five:
                df_table = pd.read_html(etree.tostring(tree))[0]
                df_table[("Unnamed: league", "league")] = (
                    df_table.xs("Comp", axis=1, level=1).squeeze().map(BIG_FIVE_DICT)
                )
                df_table[("Unnamed: season", "season")] = skey
                df_table.drop("Comp", axis=1, level=1, inplace=True)
            else:
                el = tree.xpath(f"//comment()[contains(.,'div_stats_{stat_type}')]")[0]
                df_table = pd.read_html(el.text, attrs={"id": f"stats_{stat_type}"})[0]
                df_table[("Unnamed: league", "league")] = lkey
                df_table[("Unnamed: season", "season")] = skey

            if not ("Unnamed: 2_level_0", "Nation") in df_table.columns:
                df_table.loc[:, (slice(None), "Squad")] = (
                    df_table.xs("Squad", axis=1, level=1)
                    .squeeze()
                    .apply(
                        lambda x: x.split(" ")[1] if isinstance(x, str) and x != "Squad" else None
                    )
                )
                df_table.insert(
                    2,
                    ("Unnamed: nation", "Nation"),
                    df_table.xs("Squad", axis=1, level=1).squeeze(),
                )
            else:
                df_table.loc[:, (slice(None), "Nation")] = (
                    df_table.xs("Nation", axis=1, level=1)
                    .squeeze()
                    .apply(
                        lambda x: x.split(" ")[1] if isinstance(x, str) and x != "Nation" else None
                    )
                )

            players.append(df_table)

        # return dataframe
        df = _concat(players)
        df = df[df.Player != "Player"]
        df = (
            df.drop("Matches", axis=1, level=0)
            .drop("Rk", axis=1, level=0)
            .rename(columns={"Player": "player", "Squad": "team"})
            .replace({"team": TEAMNAME_REPLACEMENTS})
            .set_index(["league", "season", "team", "player"])
            .sort_index()
        )

        return df

    def read_schedule(self, force_cache: bool = False) -> pd.DataFrame:
        """Retrieve the game schedule for the selected leagues and seasons.

        Parameters
        ----------
        force_cache : bool
             By default no cached data is used for the current season.
             If True, will force the use of cached data anyway.

        Returns
        -------
        pd.DataFrame
        """
        # get league IDs
        seasons = self.read_seasons()

        # collect teams
        schedule = []
        for (lkey, skey), season in seasons.iterrows():
            # read html page (league overview)
            url_stats = FBREF_API + season.url
            filepath_stats = self.data_dir / f"teams_{lkey}_{skey}.html"
            reader = self.get(url_stats, filepath_stats)
            tree = html.parse(reader)

            url_fixtures = FBREF_API + tree.xpath("//a[text()='Scores & Fixtures']")[0].get("href")
            filepath_fixtures = self.data_dir / f"schedule_{lkey}_{skey}.html"
            current_season = not self._is_complete(lkey, skey)
            reader = self.get(
                url_fixtures, filepath_fixtures, no_cache=current_season and not force_cache
            )
            tree = html.parse(reader)
            table = tree.xpath("//table[contains(@id, 'sched')]")[0]
            df_table = pd.read_html(etree.tostring(table))[0]
            df_table["Match Report"] = [
                mlink.xpath("./a/@href")[0]
                if mlink.xpath("./a") and mlink.xpath("./a")[0].text == "Match Report"
                else None
                for mlink in table.xpath(".//td[@data-stat='match_report']")
            ]
            df_table["league"] = lkey
            df_table["season"] = skey
            df_table = df_table.dropna(how="all")
            schedule.append(df_table)
        df = (
            pd.concat(schedule)
            .rename(
                columns={
                    "Wk": "week",
                    "Home": "home_team",
                    "Away": "away_team",
                    "xG": "home_xg",
                    "xG.1": "away_xg",
                }
            )
            .replace(
                {
                    "home_team": TEAMNAME_REPLACEMENTS,
                    "away_team": TEAMNAME_REPLACEMENTS,
                }
            )
            .pipe(standardize_colnames)
        )
        df["date"] = pd.to_datetime(df["date"]).ffill()
        df["game"] = df.apply(make_game_id, axis=1)
        df.loc[~df.match_report.isna(), "game_id"] = (
            df.loc[~df.match_report.isna(), "match_report"].str.split("/").str[3]
        )
        df = df.set_index(["league", "season", "game"]).sort_index()
        return df

    def _parse_teams(self, tree: etree.ElementTree) -> List[Dict]:
        """Parse the teams from a match summary page.

        Parameters
        ----------
        tree : etree.ElementTree
            The match summary page.

        Returns
        -------
        list of dict
        """
        team_nodes = tree.xpath("//div[@class='scorebox']//strong/a")[:2]
        teams = []
        for team in team_nodes:
            teams.append({"id": team.get("href").split("/")[3], "name": team.text.strip()})
        return teams

    def read_lineup(
        self, match_id: Optional[Union[str, List[str]]] = None, force_cache: bool = False
    ) -> pd.DataFrame:
        """Retrieve lineups for the selected leagues and seasons.

        Parameters
        ----------
        match_id : int or list of int, optional
            Retrieve the lineup for a specific game.
        force_cache : bool
            By default no cached data is used to scrape the list of available
            games for the current season. If True, will force the use of
            cached data anyway.

        Raises
        ------
        ValueError
            If no games with the given IDs were found for the selected seasons and leagues.

        Returns
        -------
        pd.DataFrame.
        """
        urlmask = FBREF_API + "/en/matches/{}"
        filemask = "match_{}.html"

        # Retrieve games for which a match report is available
        df_schedule = self.read_schedule(force_cache).reset_index()
        df_schedule = df_schedule[~df_schedule.game_id.isna() & ~df_schedule.match_report.isnull()]
        # Select requested games if available
        if match_id is not None:
            iterator = df_schedule[
                df_schedule.game_id.isin([match_id] if isinstance(match_id, str) else match_id)
            ]
            if len(iterator) == 0:
                raise ValueError("No games found with the given IDs in the selected seasons.")
        else:
            iterator = df_schedule

        lineups = []
        for i, game in iterator.iterrows():
            url = urlmask.format(game["game_id"])
            # get league and season
            logger.info(
                "[%s/%s] Retrieving game with id=%s", i + 1, len(iterator), game["game_id"]
            )
            filepath = self.data_dir / filemask.format(game["game_id"])
            reader = self.get(url, filepath)
            tree = html.parse(reader)
            teams = self._parse_teams(tree)
            tables = tree.xpath("//div[@class='lineup']")
            for i, table in enumerate(tables):
                df_table = pd.read_html(etree.tostring(table))[0]
                df_table.columns = ["jersey_number", "player"]
                df_table["team"] = teams[i]["name"]
                if "Bench" in df_table.jersey_number.values:
                    bench_idx = df_table.index[df_table.jersey_number == "Bench"][0]
                    df_table.loc[:bench_idx, "is_starter"] = True
                    df_table.loc[bench_idx:, "is_starter"] = False
                    df_table["game"] = game["game"]
                    df_table["league"] = game["league"]
                    df_table["season"] = game["season"]
                    df_table["game"] = game["game"]
                    df_table.drop(bench_idx, inplace=True)
                lineups.append(df_table)
        df = pd.concat(lineups).set_index(["league", "season", "game", "team", "player"])
        # TODO: sub in, sub out, position
        return df

    def read_player_match_stats(
        self,
        stat_type: str = "summary",
        match_id: Optional[Union[str, List[str]]] = None,
        force_cache: bool = False,
    ) -> pd.DataFrame:
        """Retrieve the match stats for the selected leagues and seasons.

        The following stat types are available:
            * 'summary'
            * 'keepers'
            * 'passing'
            * 'passing_types'
            * 'defense'
            * 'possession'
            * 'misc'

        Parameters
        ----------
        stat_type : str
            Type of stats to retrieve.
        match_id : int or list of int, optional
            Retrieve the event stream for a specific game.
        force_cache : bool
            By default no cached data is used to scrape the list of available
            games for the current season. If True, will force the use of
            cached data anyway.

        Raises
        ------
        ValueError
            If no games with the given IDs were found for the selected seasons and leagues.
        TypeError
            If ``stat_type`` is not valid.

        Returns
        -------
        pd.DataFrame
        """
        match_stats = [
            "summary",
            "keepers",
            "passing",
            "passing_types",
            "defense",
            "possession",
            "misc",
        ]

        urlmask = FBREF_API + "/en/matches/{}"
        filemask = "match_{}.html"

        if stat_type not in match_stats:
            raise TypeError(f"Invalid argument: stat_type should be in {match_stats}")

        # Retrieve games for which a match report is available
        df_schedule = self.read_schedule(force_cache).reset_index()
        df_schedule = df_schedule[~df_schedule.game_id.isna() & ~df_schedule.match_report.isnull()]
        # Selec requested games if available
        if match_id is not None:
            iterator = df_schedule[
                df_schedule.game_id.isin([match_id] if isinstance(match_id, str) else match_id)
            ]
            if len(iterator) == 0:
                raise ValueError("No games found with the given IDs in the selected seasons.")
        else:
            iterator = df_schedule

        stats = []
        for i, game in iterator.iterrows():
            url = urlmask.format(game["game_id"])
            # get league and season
            logger.info(
                "[%s/%s] Retrieving game with id=%s", i + 1, len(iterator), game["game_id"]
            )
            filepath = self.data_dir / filemask.format(game["game_id"])
            reader = self.get(url, filepath)
            tree = html.parse(reader)
            (home_team, away_team) = self._parse_teams(tree)
            if stat_type == "keepers":
                id_format = "keeper_stats_{}"
            else:
                id_format = "stats_{}_" + stat_type
            table = tree.xpath("//table[@id='" + id_format.format(home_team["id"]) + "']")[0]
            df_table = pd.read_html(etree.tostring(table))[0]
            df_table["team"] = home_team["name"]
            df_table["game"] = game["game"]
            df_table["league"] = game["league"]
            df_table["season"] = game["season"]
            df_table["game_id"] = game["game_id"]
            stats.append(df_table)
            table = tree.xpath("//table[@id='" + id_format.format(away_team["id"]) + "']")[0]
            df_table = pd.read_html(etree.tostring(table))[0]
            df_table["team"] = away_team["name"]
            df_table["game"] = game["game"]
            df_table["league"] = game["league"]
            df_table["season"] = game["season"]
            df_table["game_id"] = game["game_id"]
            stats.append(df_table)

        df = _concat(stats)
        df = df[~df.Player.str.contains(r"^\d+\sPlayers$")]
        df = (
            df.rename(columns={"Player": "player"})
            .replace({"team": TEAMNAME_REPLACEMENTS})
            .set_index(["league", "season", "game", "team", "player"])
            .sort_index()
        )
        return df

    def read_shot_events(
        self, match_id: Optional[Union[str, List[str]]] = None, force_cache: bool = False
    ) -> pd.DataFrame:
        """Retrieve shooting data for the selected seasons or selected matches.

        The data returned includes who took the shot, when, with which body
        part and from how far away. Additionally, the player creating the
        chance and also the creation before this are included in the data.

        Parameters
        ----------
        match_id : int or list of int, optional
            Retrieve the lineup for a specific game.
        force_cache : bool
            By default no cached data is used to scrape the list of available
            games for the current season. If True, will force the use of
            cached data anyway.

        Raises
        ------
        ValueError
            If no games with the given IDs were found for the selected seasons and leagues.

        Returns
        -------
        pd.DataFrame.
        """
        urlmask = FBREF_API + "/en/matches/{}"
        filemask = "match_{}.html"

        # Retrieve games for which a match report is available
        df_schedule = self.read_schedule(force_cache).reset_index()
        df_schedule = df_schedule[~df_schedule.game_id.isna() & ~df_schedule.match_report.isnull()]
        # Selec requested games if available
        if match_id is not None:
            iterator = df_schedule[
                df_schedule.game_id.isin([match_id] if isinstance(match_id, str) else match_id)
            ]
            if len(iterator) == 0:
                raise ValueError("No games found with the given IDs in the selected seasons.")
        else:
            iterator = df_schedule

        shots = []
        for i, game in iterator.iterrows():
            url = urlmask.format(game["game_id"])
            # get league and season
            logger.info(
                "[%s/%s] Retrieving game with id=%s", i + 1, len(iterator), game["game_id"]
            )
            filepath = self.data_dir / filemask.format(game["game_id"])
            reader = self.get(url, filepath)
            tree = html.parse(reader)
            df_table = pd.read_html(etree.tostring(tree), attrs={"id": "shots_all"})[0]
            df_table["league"] = game["league"]
            df_table["season"] = game["season"]
            df_table["game"] = game["game"]
            shots.append(df_table)

        df = (
            _concat(shots)
            .rename(columns={"Squad": "team"})
            .replace({"team": TEAMNAME_REPLACEMENTS})
            .pipe(
                standardize_colnames,
                cols=["Outcome", "Minute", "Distance", "Player", "Body Part", "Notes", "Event"],
            )
            .set_index(["league", "season", "game", "team", "player"])
            .sort_index()
            .dropna(how="all")
        )
        return df


def _concat(dfs: List[pd.DataFrame]) -> pd.DataFrame:
    """Merge matching tables scraped from different pages.

    The level 0 headers are not consitent across seasons and leagues, this
    function tries to determine uniform column names.

    Parameters
    ----------
    dfs : list(pd.DataFrame)
        Input dataframes.

    Returns
    -------
    pd.DataFrame
        Concatenated dataframe with uniform column names.
    """
    # Look for the most complete level 0 columns
    all_columns = []
    for df in dfs:
        columns = pd.DataFrame(df.columns.tolist())
        # Move missing columns to level 0
        columns.replace({"": None}, inplace=True)
        mask = pd.isnull(columns[1])
        columns.loc[mask, [0, 1]] = columns.loc[mask, [1, 0]].values
        # Rename unnamed columns
        mask = columns[0].str.startswith("Unnamed:").fillna(False)
        columns.loc[mask, 0] = None
        all_columns.append(columns)
    columns = reduce(lambda l, r: l.combine_first(r), all_columns)

    # Move the remaining missing columns back to level 1 and replace with empyt string
    mask = pd.isnull(columns[0])
    columns.loc[mask, [0, 1]] = columns.loc[mask, [1, 0]].values
    columns.loc[mask, 1] = ""

    for df in dfs:
        df.columns = pd.MultiIndex.from_tuples(columns.to_records(index=False).tolist())

    return pd.concat(dfs)



"""A collection of tools to read and process soccer data from various sources."""

__version__ = '1.3.2'

__all__ = [
    'FiveThirtyEight',
    'ClubElo',
    'MatchHistory',
    'FBref',
    'ESPN',
    'WhoScored',
    'SoFIFA',
]

from .clubelo import ClubElo
from .espn import ESPN
from .fbref import FBref
from .fivethirtyeight import FiveThirtyEight
from .match_history import MatchHistory
from .sofifa import SoFIFA
from .whoscored import WhoScored



"""Configurations."""

import json
import logging
import logging.config
import os
import sys
from pathlib import Path

import pretty_errors  # NOQA: F401 (imported but unused)
from rich.logging import RichHandler

# Configuration
NOCACHE = os.environ.get("SOCCERDATA_NOCACHE", 'False').lower() in ('true', '1', 't')
NOSTORE = os.environ.get("SOCCERDATA_NOSTORE", 'False').lower() in ('true', '1', 't')
LOGLEVEL = os.environ.get('SOCCERDATA_LOGLEVEL', 'INFO').upper()

# Directories
BASE_DIR = Path(os.environ.get("SOCCERDATA_DIR", Path.home() / "soccerdata"))
LOGS_DIR = Path(BASE_DIR, "logs")
DATA_DIR = Path(BASE_DIR, "data")
CONFIG_DIR = Path(BASE_DIR, "config")

# Create dirs
LOGS_DIR.mkdir(parents=True, exist_ok=True)
DATA_DIR.mkdir(parents=True, exist_ok=True)
CONFIG_DIR.mkdir(parents=True, exist_ok=True)

# Logger
logging_config = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "minimal": {"format": "%(message)s"},
        "detailed": {
            "format": "%(levelname)s %(asctime)s [%(filename)s:%(funcName)s:%(lineno)d]\n%(message)s\n"  # noqa: E501
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "stream": sys.stdout,
            "formatter": "minimal",
            "level": logging.DEBUG,
        },
        "info": {
            "class": "logging.handlers.RotatingFileHandler",
            "filename": Path(LOGS_DIR, "info.log"),
            "maxBytes": 10485760,  # 1 MB
            "backupCount": 10,
            "formatter": "detailed",
            "level": logging.INFO,
        },
        "error": {
            "class": "logging.handlers.RotatingFileHandler",
            "filename": Path(LOGS_DIR, "error.log"),
            "maxBytes": 10485760,  # 1 MB
            "backupCount": 10,
            "formatter": "detailed",
            "level": logging.ERROR,
        },
    },
    "loggers": {
        "root": {
            "handlers": ["console", "info", "error"],
            "level": LOGLEVEL,
            "propagate": True,
        },
    },
}
logging.config.dictConfig(logging_config)
logger = logging.getLogger("root")
logger.handlers[0] = RichHandler(markup=True)

# Team name replacements
TEAMNAME_REPLACEMENTS = {}
_f_custom_teamnname_replacements = CONFIG_DIR / "teamname_replacements.json"
if _f_custom_teamnname_replacements.is_file():
    with open(_f_custom_teamnname_replacements, encoding='utf8') as json_file:
        for team, to_replace_list in json.load(json_file).items():
            for to_replace in to_replace_list:
                TEAMNAME_REPLACEMENTS[to_replace] = team
    logger.info("Custom team name replacements loaded from %s.", _f_custom_teamnname_replacements)
else:
    logger.info(
        "No custom team name replacements found. You can configure these in %s.",
        _f_custom_teamnname_replacements,
    )


# League dict
LEAGUE_DICT = {
    "ENG-Premier League": {
        "ClubElo": "ENG_1",
        "MatchHistory": "E0",
        "FiveThirtyEight": "premier-league",
        "FBref": "Premier League",
        "ESPN": "eng.1",
        "SoFIFA": "English Premier League (1)",
        "WhoScored": "England - Premier League",
        "season_start": "Aug",
        "season_end": "May",
    },
    "ESP-La Liga": {
        "ClubElo": "ESP_1",
        "MatchHistory": "SP1",
        "FiveThirtyEight": "la-liga",
        "FBref": "La Liga",
        "ESPN": "esp.1",
        "SoFIFA": "Spain Primera Division (1)",
        "WhoScored": "Spain - LaLiga",
        "season_start": "Aug",
        "season_end": "May",
    },
    "ITA-Serie A": {
        "ClubElo": "ITA_1",
        "MatchHistory": "I1",
        "FiveThirtyEight": "serie-a",
        "FBref": "Serie A",
        "ESPN": "ita.1",
        "SoFIFA": " Italian Serie A (1)",
        "WhoScored": "Italy - Serie A",
        "season_start": "Aug",
        "season_end": "May",
    },
    "GER-Bundesliga": {
        "ClubElo": "GER_1",
        "MatchHistory": "D1",
        "FiveThirtyEight": "bundesliga",
        "FBref": "Fußball-Bundesliga",
        "ESPN": "ger.1",
        "SoFIFA": "German 1. Bundesliga (1)",
        "WhoScored": "Germany - Bundesliga",
        "season_start": "Aug",
        "season_end": "May",
    },
    "FRA-Ligue 1": {
        "ClubElo": "FRA_1",
        "MatchHistory": "F1",
        "FiveThirtyEight": "ligue-1",
        "FBref": "Ligue 1",
        "ESPN": "fra.1",
        "SoFIFA": "French Ligue 1 (1)",
        "WhoScored": "France - Ligue 1",
        "season_start": "Aug",
        "season_end": "May",
    },
    "INT-World Cup": {
        "FBref": "FIFA World Cup",
        "WhoScored": "International - FIFA World Cup",
    },
}
_f_custom_league_dict = CONFIG_DIR / "league_dict.json"
if _f_custom_league_dict.is_file():
    with open(_f_custom_league_dict, encoding='utf8') as json_file:
        LEAGUE_DICT = {**LEAGUE_DICT, **json.load(json_file)}
    logger.info("Custom league dict loaded from %s.", _f_custom_league_dict)
else:
    logger.info(
        "No custom league dict found. You can configure additional leagues in %s.",
        _f_custom_league_dict,
    )



"""Scraper for http://whoscored.com."""
import itertools
import json
import time
from datetime import datetime
from pathlib import Path
from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from lxml import html
from selenium.common.exceptions import (
    ElementClickInterceptedException,
    NoSuchElementException,
    TimeoutException,
)
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as ec
from selenium.webdriver.support.ui import WebDriverWait

from ._common import (
    BaseSeleniumReader,
    make_game_id,
    season_code,
    standardize_colnames,
)
from ._config import DATA_DIR, NOCACHE, NOSTORE, TEAMNAME_REPLACEMENTS, logger

WHOSCORED_DATADIR = DATA_DIR / "WhoScored"
WHOSCORED_URL = "https://www.whoscored.com"

COLS_EVENTS = {
    # The ID of the game
    "game_id": np.nan,
    # 'PreMatch', 'FirstHalf', 'SecondHalf', 'PostGame'
    "period": np.nan,
    # Integer indicating the minute of the event, ignoring stoppage time
    "minute": -1,
    # Integer indicating the second of the event, ignoring stoppage time
    "second": -1,
    # Integer indicating the minute of the event, taking into account stoppage time
    "expanded_minute": -1,
    # String describing the event type (e.g. 'Goal', 'Yellow Card', etc.)
    "type": np.nan,
    # String describing the event outcome ('Succesful' or 'Unsuccessful')
    "outcome_type": np.nan,
    # The ID of the team that the event is associated with
    "team_id": np.nan,
    # The name of the team that the event is associated with
    "team": np.nan,
    # The ID of the player that the event is associated with
    "player_id": np.nan,
    # The name of the player that the event is associated with
    "player": np.nan,
    # Coordinates of the event's location
    "x": np.nan,
    "y": np.nan,
    "end_x": np.nan,
    "end_y": np.nan,
    # Coordinates of a shot's location
    "goal_mouth_y": np.nan,
    "goal_mouth_z": np.nan,
    # The coordinates where the ball was blocked
    "blocked_x": np.nan,
    "blocked_y": np.nan,
    # List of dicts with event qualifiers
    "qualifiers": [],
    # Some boolean flags
    "is_touch": False,
    "is_shot": False,
    "is_goal": False,
    # 'Yellow', 'Red', 'SecondYellow'
    "card_type": np.nan,
    # The ID of an associated event
    "related_event_id": np.nan,
    # The ID of a secondary player that the event is associated with
    "related_player_id": np.nan,
}


class WhoScored(BaseSeleniumReader):
    """Provides pd.DataFrames from data available at http://whoscored.com.

    Data will be downloaded as necessary and cached locally in
    ``~/soccerdata/data/WhoScored``.

    Parameters
    ----------
    leagues : string or iterable, optional
        IDs of Leagues to include.
    seasons : string, int or list, optional
        Seasons to include. Supports multiple formats.
        Examples: '16-17'; 2016; '2016-17'; [14, 15, 16]
    proxy : 'tor' or dict or list(dict) or callable, optional
        Use a proxy to hide your IP address. Valid options are:
            - "tor": Uses the Tor network. Tor should be running in
              the background on port 9050.
            - dict: A dictionary with the proxy to use. The dict should be
              a mapping of supported protocols to proxy addresses. For example::

                  {
                      'http': 'http://10.10.1.10:3128',
                      'https': 'http://10.10.1.10:1080',
                  }

            - list(dict): A list of proxies to choose from. A different proxy will
              be selected from this list after failed requests, allowing rotating
              proxies.
            - callable: A function that returns a valid proxy. This function will
              be called after failed requests, allowing rotating proxies.
    no_cache : bool
        If True, will not use cached data.
    no_store : bool
        If True, will not store downloaded data.
    data_dir : Path
        Path to directory where data will be cached.
    path_to_browser : Path, optional
        Path to the Chrome executable.
    headless : bool, default: True
        If True, will run Chrome in headless mode. Setting this to False might
        help to avoid getting blocked.
    """

    def __init__(
        self,
        leagues: Optional[Union[str, List[str]]] = None,
        seasons: Optional[Union[str, int, Iterable[Union[str, int]]]] = None,
        proxy: Optional[
            Union[str, Dict[str, str], List[Dict[str, str]], Callable[[], Dict[str, str]]]
        ] = None,
        no_cache: bool = NOCACHE,
        no_store: bool = NOSTORE,
        data_dir: Path = WHOSCORED_DATADIR,
        path_to_browser: Optional[Path] = None,
        headless: bool = False,
    ):
        """Initialize the WhoScored reader."""
        super().__init__(
            leagues=leagues,
            proxy=proxy,
            no_cache=no_cache,
            no_store=no_store,
            data_dir=data_dir,
            path_to_browser=path_to_browser,
            headless=headless,
        )
        self.seasons = seasons  # type: ignore
        self.rate_limit = 5
        self.max_delay = 5
        if not self.no_store:
            (self.data_dir / "seasons").mkdir(parents=True, exist_ok=True)
            (self.data_dir / "matches").mkdir(parents=True, exist_ok=True)
            (self.data_dir / "previews").mkdir(parents=True, exist_ok=True)
            (self.data_dir / "events").mkdir(parents=True, exist_ok=True)

    def read_leagues(self) -> pd.DataFrame:
        """Retrieve the selected leagues from the datasource.

        Returns
        -------
        pd.DataFrame
        """
        url = WHOSCORED_URL
        filepath = self.data_dir / "tiers.json"
        reader = self.get(url, filepath, var="allRegions")

        data = json.load(reader)

        leagues = []
        for region in data:
            for league in region["tournaments"]:
                leagues.append(
                    {
                        "region_id": region["id"],
                        "region": region["name"],
                        "league_id": league["id"],
                        "league": league["name"],
                        "url": league["url"],
                    }
                )

        df = (
            pd.DataFrame(leagues)
            .assign(league=lambda x: x.region + " - " + x.league)
            .pipe(self._translate_league)
            .set_index("league")
            .loc[self._selected_leagues.keys()]
            .sort_index()
        )
        return df

    def read_seasons(self) -> pd.DataFrame:
        """Retrieve the selected seasons for the selected leagues.

        Returns
        -------
        pd.DataFrame
        """
        df_leagues = self.read_leagues()

        seasons = []
        for lkey, league in df_leagues.iterrows():
            url = WHOSCORED_URL + league.url
            filemask = "seasons/{}.html"
            filepath = self.data_dir / filemask.format(lkey)
            reader = self.get(url, filepath, var=None)

            # extract team links
            tree = html.parse(reader)
            for node in tree.xpath("//select[contains(@id,'seasons')]/option"):
                # extract team IDs from links
                seasons.append(
                    {
                        "url": node.get("value"),
                        "league": lkey,
                        "league_id": league.league_id,
                        "season": season_code(node.text),
                    }
                )

        df = (
            pd.DataFrame(seasons)
            .set_index(["league", "season"])
            .sort_index()
            .loc[itertools.product(self.leagues, self.seasons)]
        )
        return df

    def _parse_season_stages(self) -> List[Dict]:
        match_selector = (
            "//div[contains(@id,'tournament-fixture')]//div[contains(@class,'divtable-row')]"
        )
        WebDriverWait(self._driver, 30, poll_frequency=1).until(
            ec.presence_of_element_located((By.XPATH, match_selector))
        )
        node_stages_selector = "//select[contains(@id,'stages')]/option"
        node_stages = self._driver.find_elements(By.XPATH, node_stages_selector)
        stages = []
        for stage in node_stages:
            stages.append({"url": stage.get_attribute("value"), "name": stage.text})
        return stages

    def _parse_schedule_page(self) -> Tuple[List[Dict], Optional[WebElement]]:
        match_selector = (
            "//div[contains(@id,'tournament-fixture')]//div[contains(@class,'divtable-row')]"
        )
        date_selector = "./div[contains(@class,'divtable-header')]"
        time_selector = "./div[contains(@class,'time')]"
        home_team_selector = "./div[contains(@class,'team home')]//a"
        away_team_selector = "./div[contains(@class,'team away')]//a"
        result_selector = "./div[contains(@class,'result')]//a"

        try:
            WebDriverWait(self._driver, 30, poll_frequency=1).until(
                ec.presence_of_element_located((By.XPATH, match_selector))
            )
            date_str = None
            schedule_page = []
            for node in self._driver.find_elements(By.XPATH, match_selector):
                if node.get_attribute("data-id"):
                    match_id = int(node.get_attribute("data-id"))
                    time_str = node.find_element(By.XPATH, time_selector).get_attribute(
                        "textContent"
                    )
                    match_url = node.find_element(By.XPATH, result_selector).get_attribute("href")
                    schedule_page.append(
                        {
                            "date": datetime.strptime(
                                f"{date_str} {time_str}", "%A, %b %d %Y %H:%M"
                            ),
                            "home_team": node.find_element(By.XPATH, home_team_selector).text,
                            "away_team": node.find_element(By.XPATH, away_team_selector).text,
                            "game_id": match_id,
                            "url": match_url,
                        }
                    )
                else:
                    date_str = node.find_element(By.XPATH, date_selector).text
                    logger.info("Scraping game schedule for %s", date_str)
        except TimeoutException:
            schedule_page = []

        try:
            next_page_selector = (
                "//div[contains(@id,'date-controller')]"
                "/a[contains(@class,'previous') and not(contains(@class, 'is-disabled'))]"
            )
            next_page = self._driver.find_element(By.XPATH, next_page_selector)
        except NoSuchElementException:
            next_page = None
        return schedule_page, next_page

    def _parse_schedule(self, stage: Optional[str] = None) -> List[Dict]:
        schedule = []
        # Parse first page
        page_schedule, next_page = self._parse_schedule_page()
        schedule.extend(page_schedule)
        # Go to next page
        while next_page is not None:
            try:
                next_page.click()
                time.sleep(5)
                logger.debug("Next page")
            except ElementClickInterceptedException:
                self._handle_banner()
            # Parse next page
            page_schedule, next_page = self._parse_schedule_page()
            schedule.extend(page_schedule)
        schedule = [dict(item, stage=stage) for item in schedule]
        return schedule

    def read_schedule(self, force_cache: bool = False) -> pd.DataFrame:
        """Retrieve the game schedule for the selected leagues and seasons.

        Parameters
        ----------
        force_cache : bool
             By default no cached data is used for the current season.
             If True, will force the use of cached data anyway.

        Returns
        -------
        pd.DataFrame
        """
        df_seasons = self.read_seasons()
        filemask = "matches/{}_{}.csv"

        all_schedules = []
        for (lkey, skey), season in df_seasons.iterrows():

            filepath = self.data_dir / filemask.format(lkey, skey)
            url = WHOSCORED_URL + season.url

            schedule = []
            is_current_season = not self._is_complete(lkey, skey)
            no_cache = (not filepath.exists()) or self.no_cache
            if (is_current_season and not force_cache) or no_cache:
                # Scrape the season's schedule
                self._driver.get(url)

                # Check if season consists of multiple stages
                stages = self._parse_season_stages()

                # Handle a multi-stage season
                if len(stages) > 0:
                    for stage in stages:
                        url = WHOSCORED_URL + stage["url"].replace("Show", "Fixtures")
                        self._driver.get(url)
                        try:
                            WebDriverWait(self._driver, 30, poll_frequency=1).until(
                                ec.presence_of_element_located(
                                    (By.XPATH, "//div[@id='tournament-fixture']")
                                )
                            )
                        except TimeoutException:
                            # Tournaments sometimes do not have a fixtures page,
                            # the summary page has to be used instead
                            url = WHOSCORED_URL + stage["url"]
                            self._driver.get(url)
                        logger.info("Scraping game schedule with stage=%s from %s", stage, url)
                        schedule.extend(self._parse_schedule(stage=stage["name"]))

                # Handle a single-stage season
                else:
                    fixtures_nav_selector = "//a[text()='Fixtures']"
                    fixtures_nav = self._driver.find_element(By.XPATH, fixtures_nav_selector)
                    self._driver.get(fixtures_nav.get_attribute("href"))
                    try:
                        WebDriverWait(self._driver, 30, poll_frequency=1).until(
                            ec.presence_of_element_located(
                                (By.XPATH, "//div[@id='tournament-fixture']")
                            )
                        )
                    except TimeoutException:
                        # Tournaments sometimes do not have a fixtures page,
                        # the summary page has to be used instead
                        summary_nav_selector = "//a[text()='Fixtures']"
                        summary_nav = self._driver.find_element(By.XPATH, summary_nav_selector)
                        self._driver.get(summary_nav.get_attribute("href"))
                    logger.info("Scraping game schedule from %s", url)
                    schedule.extend(self._parse_schedule())

                # Cache the data
                df_schedule = pd.DataFrame(schedule).assign(league=lkey, season=skey)
                if not self.no_store:
                    df_schedule.to_csv(filepath, index=False)

            else:
                # Load cached data
                logger.info("Retrieving game schedule of %s - %s from the cache", lkey, skey)
                df_schedule = pd.read_csv(filepath)

            all_schedules.append(df_schedule)

        # Construct the output dataframe
        df = (
            pd.concat(all_schedules)
            .drop_duplicates()
            .replace(
                {
                    "home_team": TEAMNAME_REPLACEMENTS,
                    "away_team": TEAMNAME_REPLACEMENTS,
                }
            )
            .assign(date=lambda x: pd.to_datetime(x["date"]))
            .assign(game=lambda df: df.apply(make_game_id, axis=1))
            .set_index(["league", "season", "game"])
            .sort_index()
        )
        return df

    def _read_game_info(self, game_id: int) -> Dict:
        """Return game info available in the header."""
        urlmask = WHOSCORED_URL + "/Matches/{}"
        url = urlmask.format(game_id)
        data = {}
        self._driver.get(url)
        # league and season
        breadcrumb = self._driver.find_elements(
            By.XPATH, "//div[@id='breadcrumb-nav']/*[not(contains(@class, 'separator'))]"
        )
        country = breadcrumb[0].text
        league, season = breadcrumb[1].text.split(" - ")
        data["league"] = {v: k for k, v in self._all_leagues().items()}[f"{country} - {league}"]
        data["season"] = season_code(season)
        # match header
        match_header = self._driver.find_element(By.XPATH, "//div[@id='match-header']")
        score_info = match_header.find_element(By.XPATH, ".//div[@class='teams-score-info']")
        data["home_team"] = score_info.find_element(
            By.XPATH, "./span[contains(@class,'home team')]"
        ).text
        data["result"] = score_info.find_element(
            By.XPATH, "./span[contains(@class,'result')]"
        ).text
        data["away_team"] = score_info.find_element(
            By.XPATH, "./span[contains(@class,'away team')]"
        ).text
        info_blocks = match_header.find_elements(By.XPATH, ".//div[@class='info-block cleared']")
        for block in info_blocks:
            for desc_list in block.find_elements(By.TAG_NAME, "dl"):
                for desc_def in desc_list.find_elements(By.TAG_NAME, "dt"):
                    desc_val = desc_def.find_element(By.XPATH, "./following-sibling::dd")
                    data[desc_def.text] = desc_val.text

        return data

    def read_missing_players(
        self, match_id: Optional[Union[int, List[int]]] = None, force_cache: bool = False
    ) -> pd.DataFrame:
        """Retrieve a list of injured and suspended players ahead of each game.

        Parameters
        ----------
        match_id : int or list of int, optional
            Retrieve the missing players for a specific game.
        force_cache : bool
            By default no cached data is used to scrapre the list of available
            games for the current season. If True, will force the use of
            cached data anyway.

        Raises
        ------
        ValueError
            If the given match_id could not be found in the selected seasons.

        Returns
        -------
        pd.DataFrame
        """
        urlmask = WHOSCORED_URL + "/Matches/{}/Preview"
        filemask = "WhoScored/previews/{}_{}/{}.html"

        df_schedule = self.read_schedule(force_cache).reset_index()
        if match_id is not None:
            iterator = df_schedule[
                df_schedule.game_id.isin([match_id] if isinstance(match_id, int) else match_id)
            ]
            if len(iterator) == 0:
                raise ValueError("No games found with the given IDs in the selected seasons.")
        else:
            iterator = df_schedule.sample(frac=1)

        match_sheets = []
        for i, (_, game) in enumerate(iterator.iterrows()):
            url = urlmask.format(game.game_id)
            filepath = DATA_DIR / filemask.format(game["league"], game["season"], game["game_id"])

            logger.info(
                "[%s/%s] Retrieving game with id=%s", i + 1, len(iterator), game["game_id"]
            )
            reader = self.get(url, filepath, var=None)

            # extract missing players
            tree = html.parse(reader)
            for node in tree.xpath("//div[@id='missing-players']/div[2]/table/tbody/tr"):
                # extract team IDs from links
                match_sheets.append(
                    {
                        "league": game["league"],
                        "season": game["season"],
                        "game": game["game"],
                        "game_id": game["game_id"],
                        "team": game["home_team"],
                        "player": node.xpath("./td[contains(@class,'pn')]/a")[0].text,
                        "player_id": int(
                            node.xpath("./td[contains(@class,'pn')]/a")[0]
                            .get("href")
                            .split("/")[2]
                        ),
                        "reason": node.xpath("./td[contains(@class,'reason')]/span")[0].get(
                            "title"
                        ),
                        "status": node.xpath("./td[contains(@class,'confirmed')]")[0].text,
                    }
                )
            for node in tree.xpath("//div[@id='missing-players']/div[3]/table/tbody/tr"):
                # extract team IDs from links
                match_sheets.append(
                    {
                        "league": game["league"],
                        "season": game["season"],
                        "game": game["game"],
                        "game_id": game["game_id"],
                        "team": game["away_team"],
                        "player": node.xpath("./td[contains(@class,'pn')]/a")[0].text,
                        "player_id": int(
                            node.xpath("./td[contains(@class,'pn')]/a")[0]
                            .get("href")
                            .split("/")[2]
                        ),
                        "reason": node.xpath("./td[contains(@class,'reason')]/span")[0].get(
                            "title"
                        ),
                        "status": node.xpath("./td[contains(@class,'confirmed')]")[0].text,
                    }
                )
        df = (
            pd.DataFrame(match_sheets)
            .set_index(["league", "season", "game", "team", "player"])
            .sort_index()
        )
        return df

    def read_events(  # noqa: C901
        self,
        match_id: Optional[Union[int, List[int]]] = None,
        force_cache: bool = False,
        live: bool = False,
        output_fmt: Optional[str] = "events",
    ) -> Optional[Union[pd.DataFrame, Dict[int, List], "OptaLoader"]]:  # type: ignore  # noqa: F821
        """Retrieve the the event data for each game in the selected leagues and seasons.

        Parameters
        ----------
        match_id : int or list of int, optional
            Retrieve the event stream for a specific game.
        force_cache : bool
            By default no cached data is used to scrape the list of available
            games for the current season. If True, will force the use of
            cached data anyway.
        live : bool
            If True, will not return a cached copy of the event data. This is
            usefull to scrape live data.
        output_fmt : str, default: 'events'
            The output format of the returned data. Possible values are:
                - 'events' (default): Returns a dataframe with all events.
                - 'raw': Returns the original unformatted WhoScored JSON.
                - 'spadl': Returns a dataframe with the SPADL representation
                  of the original events.
                  See https://socceraction.readthedocs.io/en/latest/documentation/SPADL.html#spadl
                - 'atomic-spadl': Returns a dataframe with the Atomic-SPADL representation
                  of the original events.
                  See https://socceraction.readthedocs.io/en/latest/documentation/SPADL.html#atomic-spadl
                - 'loader': Returns a socceraction.data.opta.OptaLoader
                  instance, which can be used to retrieve the actual data.
                  See https://socceraction.readthedocs.io/en/latest/modules/generated/socceraction.data.opta.OptaLoader.html#socceraction.data.opta.OptaLoader  # noqa: E501
                - None: Doesn't return any data. This is useful to just cache
                  the data without storing the events in memory.

        Raises
        ------
        ValueError
            If the given match_id could not be found in the selected seasons.
        ImportError
            If the requested output format is 'spadl', 'atomic-spadl' or
            'loader' but the socceraction package is not installed.

        Returns
        -------
        See the description of the ``output_fmt`` parameter.
        """
        output_fmt = output_fmt.lower() if output_fmt is not None else None
        if output_fmt in ["loader", "spadl", "atomic-spadl"]:
            if self.no_store:
                raise ValueError(
                    f"The '{output_fmt}' output format is not supported "
                    "when using the 'no_store' option."
                )
            try:
                from socceraction.atomic.spadl import convert_to_atomic
                from socceraction.data.opta import OptaLoader
                from socceraction.data.opta.loader import _eventtypesdf
                from socceraction.data.opta.parsers import WhoScoredParser
                from socceraction.spadl.opta import convert_to_actions

                if output_fmt == "loader":
                    import socceraction
                    from packaging import version

                    if version.parse(socceraction.__version__) < version.parse("1.2.3"):
                        raise ImportError(
                            "The 'loader' output format requires socceraction >= 1.2.3"
                        )
            except ImportError:
                raise ImportError(
                    "The socceraction package is required to use the 'spadl' "
                    "or 'atomic-spadl' output format. "
                    "Please install it with `pip install socceraction`."
                )
        urlmask = WHOSCORED_URL + "/Matches/{}/Live"
        filemask = "events/{}_{}/{}.json"

        df_schedule = self.read_schedule(force_cache).reset_index()
        if match_id is not None:
            iterator = df_schedule[
                df_schedule.game_id.isin([match_id] if isinstance(match_id, int) else match_id)
            ]
            if len(iterator) == 0:
                raise ValueError("No games found with the given IDs in the selected seasons.")
        else:
            iterator = df_schedule.sample(frac=1)

        events = {}
        player_names = {}
        team_names = {}
        for i, (_, game) in enumerate(iterator.iterrows()):
            url = urlmask.format(game["game_id"])
            # get league and season
            logger.info(
                "[%s/%s] Retrieving game with id=%s", i + 1, len(iterator), game["game_id"]
            )
            filepath = self.data_dir / filemask.format(
                game["league"], game["season"], game["game_id"]
            )

            reader = self.get(
                url,
                filepath,
                var="requirejs.s.contexts._.config.config.params.args.matchCentreData",
                no_cache=live,
            )
            json_data = json.load(reader)
            if json_data is not None:
                player_names.update(
                    {int(k): v for k, v in json_data["playerIdNameDictionary"].items()}
                )
                team_names.update(
                    {
                        int(json_data[side]["teamId"]): json_data[side]["name"]
                        for side in ["home", "away"]
                    }
                )
                if "events" in json_data:
                    game_events = json_data["events"]
                    if output_fmt == "events":
                        df_events = pd.DataFrame(game_events)
                        df_events["game"] = game["game"]
                        df_events["league"] = game["league"]
                        df_events["season"] = game["season"]
                        df_events["game_id"] = game["game_id"]
                        events[game["game_id"]] = df_events
                    elif output_fmt == "raw":
                        events[game["game_id"]] = game_events
                    elif output_fmt in ["spadl", "atomic-spadl"]:
                        parser = WhoScoredParser(
                            str(filepath),
                            competition_id=game["league"],
                            season_id=game["season"],
                            game_id=game["game_id"],
                        )
                        df_events = (
                            pd.DataFrame.from_dict(parser.extract_events(), orient="index")
                            .merge(_eventtypesdf, on="type_id", how="left")
                            .reset_index(drop=True)
                        )
                        df_actions = convert_to_actions(
                            df_events, home_team_id=int(json_data["home"]["teamId"])
                        )
                        if output_fmt == "spadl":
                            events[game["game_id"]] = df_actions
                        else:
                            events[game["game_id"]] = convert_to_atomic(df_actions)

            else:
                logger.warning("No events found for game %s", game["game_id"])

        if output_fmt is None:
            return None

        if output_fmt == "raw":
            return events

        if output_fmt == "loader":
            return OptaLoader(
                root=self.data_dir,
                parser="whoscored",
                feeds={"whoscored": "events/{competition_id}_{season_id}/{game_id}.json"},
            )

        df = (
            pd.concat(events.values())
            .pipe(standardize_colnames)
            .assign(
                player=lambda x: x.player_id.replace(player_names),
                team=lambda x: x.team_id.replace(team_names).replace(TEAMNAME_REPLACEMENTS),
            )
        )

        if output_fmt == "events":
            df = df.set_index(["league", "season", "game", "id"]).sort_index()
            # add missing columns
            for col, default in COLS_EVENTS.items():
                if col not in df.columns:
                    df[col] = default
            df["outcome_type"] = df["outcome_type"].apply(
                lambda x: x.get("displayName") if pd.notnull(x) else x
            )
            df["card_type"] = df["card_type"].apply(
                lambda x: x.get("displayName") if pd.notnull(x) else x
            )
            df["type"] = df["type"].apply(lambda x: x.get("displayName") if pd.notnull(x) else x)
            df["period"] = df["period"].apply(
                lambda x: x.get("displayName") if pd.notnull(x) else x
            )
            df = df[list(COLS_EVENTS.keys())]

        return df

    def _handle_banner(self) -> None:
        try:
            # self._driver.get(WHOSCORED_URL)
            time.sleep(2)
            self._driver.find_element(By.XPATH, "//button[./span[text()='AGREE']]").click()
            time.sleep(2)
        except NoSuchElementException:
            with open("/tmp/error.html", "w") as f:
                f.write(self._driver.page_source)
            raise ElementClickInterceptedException()



"""Scraper for https://projects.fivethirtyeight.com/soccer-predictions."""
import itertools
import json
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

import pandas as pd

from ._common import BaseRequestsReader, make_game_id, standardize_colnames
from ._config import DATA_DIR, NOCACHE, NOSTORE, TEAMNAME_REPLACEMENTS

FIVETHIRTYEIGHT_DATA_DIR = DATA_DIR / "FiveThirtyEight"
FIVETHIRTYEIGHT_API = "https://projects.fivethirtyeight.com/soccer-predictions"


class FiveThirtyEight(BaseRequestsReader):
    """Provides pd.DataFrames from fivethirtyeight's "Club Soccer Predictions" project.

    Data will be downloaded as necessary and cached locally in
    ``~/soccerdata/data/FiveThirtyEight``.

    Original project and background info:
    https://projects.fivethirtyeight.com/soccer-predictions/ and
    https://fivethirtyeight.com/features/how-our-club-soccer-projections-work/


    Parameters
    ----------
    leagues : string or iterable, optional
        IDs of Leagues to include.
    seasons : string, int or list, optional
        Seasons to include. Supports multiple formats.
        Examples: '16-17'; 2016; '2016-17'; [14, 15, 16]
    proxy : 'tor' or dict or list(dict) or callable, optional
        Use a proxy to hide your IP address. Valid options are:
            - "tor": Uses the Tor network. Tor should be running in
              the background on port 9050.
            - dict: A dictionary with the proxy to use. The dict should be
              a mapping of supported protocols to proxy addresses. For example::

                  {
                      'http': 'http://10.10.1.10:3128',
                      'https': 'http://10.10.1.10:1080',
                  }

            - list(dict): A list of proxies to choose from. A different proxy will
              be selected from this list after failed requests, allowing rotating
              proxies.
            - callable: A function that returns a valid proxy. This function will
              be called after failed requests, allowing rotating proxies.
    no_cache : bool
        If True, will not use cached data.
    no_store : bool
        If True, will not store downloaded data.
    data_dir : Path
        Path to directory where data will be cached.
    """

    def __init__(
        self,
        leagues: Optional[Union[str, List[str]]] = None,
        seasons: Optional[Union[str, int, List]] = None,
        proxy: Optional[
            Union[str, Dict[str, str], List[Dict[str, str]], Callable[[], Dict[str, str]]]
        ] = None,
        no_cache: bool = NOCACHE,
        no_store: bool = NOSTORE,
        data_dir: Path = FIVETHIRTYEIGHT_DATA_DIR,
    ):
        """Initialize a new FiveThirtyEight reader."""
        super().__init__(
            leagues=leagues, proxy=proxy, no_cache=no_cache, no_store=no_store, data_dir=data_dir
        )
        self.seasons = seasons  # type: ignore
        self._data = {}

        url = f"{FIVETHIRTYEIGHT_API}/data.json"
        filepath = self.data_dir / "latest.json"
        reader = self.get(url, filepath)

        for k, v in json.load(reader).items():
            self._data[k] = v

    def read_leagues(self) -> pd.DataFrame:
        """Retrieve the selected leagues from the datasource.

        Returns
        -------
        pd.DataFrame
        """
        df = (
            pd.DataFrame.from_dict(self._data["leagues"])
            .rename(columns={"slug": "league", "id": "league_id"})
            .pipe(self._translate_league)
            .pipe(standardize_colnames)
            .drop(columns=["overview_column", "custom_template", "skip_cols"])
            .set_index("league")
            .loc[self._selected_leagues.keys()]
            .sort_index()
        )
        return df

    def read_games(self) -> pd.DataFrame:
        """Retrieve all games for the selected leagues.

        Returns
        -------
        pd.DataFrame
        """
        col_rename = {
            "adj_score1": "adj_score_home",
            "adj_score2": "adj_score_away",
            "chances1": "chances_home",
            "chances2": "chances_away",
            "datetime": "date",
            "moves1": "moves_home",
            "moves2": "moves_away",
            "prob1": "prob_home",
            "prob2": "prob_away",
            "probtie": "prob_tie",
            "score1": "score_home",
            "score2": "score_away",
            "team1": "home_team",
            "team1_code": "home_code",
            "team1_id": "home_id",
            "team1_sdr_id": "home_sdr_id",
            "team2": "away_team",
            "team2_code": "away_code",
            "team2_id": "away_id",
            "team2_sdr_id": "away_sdr_id",
        }

        filemask = "matches_{}_{}.csv"
        urlmask = FIVETHIRTYEIGHT_API + "/forecasts/20{}_{}_matches.json"
        data = []
        for lkey, skey in itertools.product(self._selected_leagues.values(), self.seasons):
            filepath = self.data_dir / filemask.format(lkey, skey)
            url = urlmask.format(skey[:2], lkey)
            reader = self.get(url, filepath)
            data.extend([{"league": lkey, "season": skey, **d} for d in json.load(reader)])

        df = (
            pd.DataFrame.from_dict(data)
            .rename(columns=col_rename)
            .assign(date=lambda x: pd.to_datetime(x["date"]))
            .replace(
                {
                    "home_team": TEAMNAME_REPLACEMENTS,
                    "away_team": TEAMNAME_REPLACEMENTS,
                }
            )
            .drop("id", axis=1)
            .drop("league_id", axis=1)
            .replace("None", float("nan"))
            .pipe(self._translate_league)
        )

        df = df[~df.date.isna()]
        df["game"] = df.apply(make_game_id, axis=1)
        df.set_index(["league", "season", "game"], inplace=True)
        df.sort_index(inplace=True)
        return df

    def read_forecasts(self) -> pd.DataFrame:
        """Retrieve the forecasted results for the selected leagues.

        Returns
        -------
        pd.DataFrame
        """
        filemask = "forecasts_{}_{}.csv"
        urlmask = FIVETHIRTYEIGHT_API + "/forecasts/20{}_{}_forecast.json"
        data = []
        for lkey, skey in itertools.product(self._selected_leagues.values(), self.seasons):
            filepath = self.data_dir / filemask.format(lkey, skey)
            url = urlmask.format(skey[:2], lkey)
            reader = self.get(url, filepath)

            forecasts = json.load(reader)
            for f in forecasts["forecasts"]:
                for t in f["teams"]:
                    data.append(
                        {
                            "league": lkey,
                            "season": skey,
                            "last_updated": f["last_updated"],
                            **t,
                        }
                    )
        df = (
            pd.DataFrame.from_dict(data)
            .rename(columns={"name": "team"})
            .replace({"team": TEAMNAME_REPLACEMENTS})
            .replace("None", float("nan"))
            .pipe(self._translate_league)
            .set_index(["league", "season", "last_updated", "team"])
            .sort_index()
        )
        return df

    def read_clinches(self) -> pd.DataFrame:
        """Retrieve clinches for the selected leagues.

        Returns
        -------
        pd.DataFrame
        """
        filemask = "clinches_{}_{}.csv"
        urlmask = FIVETHIRTYEIGHT_API + "/forecasts/20{}_{}_clinches.json"
        data = []
        for lkey, skey in itertools.product(self._selected_leagues.values(), self.seasons):
            filepath = self.data_dir / filemask.format(lkey, skey)
            url = urlmask.format(skey[:2], lkey)
            reader = self.get(url, filepath)
            data.extend([{"league": lkey, "season": skey, **c} for c in json.load(reader)])

        teams = (
            self.read_games()[["home_team", "home_id"]]
            .drop_duplicates()
            .rename(columns={"home_team": "team", "home_id": "team_id"})
        )
        df = (
            pd.DataFrame.from_dict(data)
            .assign(date=lambda x: pd.to_datetime(x["dt"]))
            .merge(teams, on="team_id", how="left")
            .replace({"team": TEAMNAME_REPLACEMENTS})
            .drop("dt", axis=1)
            .drop("league_id", axis=1)
            .drop("team_id", axis=1)
            .pipe(self._translate_league)
            .set_index(["league", "season", "date"])
            .sort_index()
        )
        return df



"""Scraper for http://site.api.espn.com/apis/site/v2/sports/soccer."""
import datetime
import itertools
import json
import re
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

import pandas as pd
import requests

from ._common import BaseRequestsReader, make_game_id, standardize_colnames
from ._config import DATA_DIR, NOCACHE, NOSTORE, TEAMNAME_REPLACEMENTS, logger

# http://site.api.espn.com/apis/site/v2/sports/soccer/eng.1/summary?event=513466
# http://site.api.espn.com/apis/site/v2/sports/soccer/eng.1/scoreboard?dates=20180901

ESPN_DATADIR = DATA_DIR / "ESPN"
ESPN_API = "http://site.api.espn.com/apis/site/v2/sports/soccer"


class ESPN(BaseRequestsReader):
    """Provides pd.DataFrames from JSON api available at http://site.api.espn.com.

    Data will be downloaded as necessary and cached locally in
    ``~/soccerdata/data/ESPN``.

    Parameters
    ----------
    leagues : string or iterable, optional
        IDs of leagues to include.

    seasons : string, int or list, optional
        Seasons to include. Supports multiple formats.
        Examples: '16-17'; 2016; '2016-17'; [14, 15, 16]
    proxy : 'tor' or dict or list(dict) or callable, optional
        Use a proxy to hide your IP address. Valid options are:
            - "tor": Uses the Tor network. Tor should be running in
              the background on port 9050.
            - dict: A dictionary with the proxy to use. The dict should be
              a mapping of supported protocols to proxy addresses. For example::

                  {
                      'http': 'http://10.10.1.10:3128',
                      'https': 'http://10.10.1.10:1080',
                  }

            - list(dict): A list of proxies to choose from. A different proxy will
              be selected from this list after failed requests, allowing rotating
              proxies.
            - callable: A function that returns a valid proxy. This function will
              be called after failed requests, allowing rotating proxies.
    no_cache : bool
        If True, will not use cached data.
    no_store : bool
        If True, will not store downloaded data.
    data_dir : Path
        Path to directory where data will be cached.
    """

    def __init__(
        self,
        leagues: Optional[Union[str, List[str]]] = None,
        seasons: Optional[Union[str, int, List]] = None,
        proxy: Optional[
            Union[str, Dict[str, str], List[Dict[str, str]], Callable[[], Dict[str, str]]]
        ] = None,
        no_cache: bool = NOCACHE,
        no_store: bool = NOSTORE,
        data_dir: Path = ESPN_DATADIR,
    ):
        """Initialize a new ESPN reader."""
        super().__init__(
            leagues=leagues,
            proxy=proxy,
            no_cache=no_cache,
            no_store=no_store,
            data_dir=data_dir,
        )
        self.seasons = seasons  # type: ignore

    def read_schedule(self, force_cache: bool = False) -> pd.DataFrame:
        """Retrieve the game schedule for the selected leagues and seasons.

        Parameters
        ----------
        force_cache : bool
             By default no cached data is used for the current season.
             If True, will force the use of cached data anyway.

        Returns
        -------
        pd.DataFrame
        """
        urlmask = ESPN_API + "/{}/scoreboard?dates={}"
        filemask = "Schedule_{}_{}.json"

        df_list = []
        # Get match days
        for lkey, skey in itertools.product(self._selected_leagues.values(), self.seasons):
            if int(skey[:2]) > int(str(datetime.datetime.now().year + 1)[-2:]):
                start_date = "".join(["19", skey[:2], "07", "01"])
            else:
                start_date = "".join(["20", skey[:2], "07", "01"])

            url = urlmask.format(lkey, start_date)
            resp = requests.get(url=url)
            data = resp.json()

            match_dates = [
                datetime.datetime.strptime(d, "%Y-%m-%dT%H:%MZ").strftime("%Y%m%d")
                for d in data["leagues"][0]["calendar"]
            ]
            for date in match_dates:
                url = urlmask.format(lkey, date)
                filepath = self.data_dir / filemask.format(lkey, date)
                current_season = not self._is_complete(lkey, skey)
                reader = self.get(url, filepath, no_cache=current_season and not force_cache)

                data = json.load(reader)
                df_list.extend(
                    [
                        {
                            "league": lkey,
                            "season": skey,
                            "date": e["date"],
                            "home_team": e["competitions"][0]["competitors"][0]["team"]["name"],
                            "away_team": e["competitions"][0]["competitors"][1]["team"]["name"],
                            "game_id": int(e["id"]),
                            "league_id": lkey,
                        }
                        for e in data["events"]
                    ]
                )
        df = (
            pd.DataFrame(df_list)
            .pipe(self._translate_league)
            .replace({"home_team": TEAMNAME_REPLACEMENTS, "away_team": TEAMNAME_REPLACEMENTS})
            .assign(date=lambda x: pd.to_datetime(x["date"]))
            .dropna(subset=["home_team", "away_team", "date"])
            .assign(game=lambda df: df.apply(make_game_id, axis=1))
            .set_index(["league", "season", "game"])
            .sort_index()
        )

        return df

    def read_matchsheet(self, match_id: Optional[Union[int, List[int]]] = None) -> pd.DataFrame:
        """Retrieve match sheets for the selected leagues and seasons.

        Parameters
        ----------
        match_id : int or list of int, optional
            Retrieve the match sheet for a specific game.

        Raises
        ------
        ValueError
            If no games with the given IDs were found for the selected seasons and leagues.

        Returns
        -------
        pd.DataFrame.
        """
        urlmask = ESPN_API + "/{}/summary?event={}"
        filemask = "Summary_{}.json"

        df_schedule = self.read_schedule().reset_index()
        if match_id is not None:
            iterator = df_schedule[
                df_schedule.game_id.isin([match_id] if isinstance(match_id, int) else match_id)
            ]
            if len(iterator) == 0:
                raise ValueError(
                    "No games with the given IDs found for the selected seasons and leagues."
                )
        else:
            iterator = df_schedule

        df_list = []
        for i, match in iterator.iterrows():
            url = urlmask.format(match["league_id"], match["game_id"])
            filepath = self.data_dir / filemask.format(match["game_id"])
            reader = self.get(url, filepath)

            data = json.load(reader)
            for i in range(2):
                match_sheet = {
                    "game": match["game"],
                    "league": match["league"],
                    "season": match["season"],
                    "team": data["boxscore"]["form"][i]["team"]["displayName"],
                    "is_home": (i == 0),
                    "venue": data["gameInfo"]["venue"]["fullName"]
                    if "venue" in data["gameInfo"]
                    else None,
                    "attendance": data["gameInfo"]["attendance"],
                    "capacity": data["gameInfo"]["venue"]["capacity"]
                    if "venue" in data["gameInfo"]
                    else None,
                    "roster": data["rosters"][i]["roster"]
                    if "roster" in data["rosters"][i]
                    else None,
                }
                if "statistics" in data["boxscore"]["teams"][i]:
                    for stat in data["boxscore"]["teams"][i]["statistics"]:
                        match_sheet[stat["name"]] = stat["displayValue"]
                df_list.append(match_sheet)
        df = (
            pd.DataFrame(df_list)
            .replace({"team": TEAMNAME_REPLACEMENTS})
            .pipe(standardize_colnames)
            .set_index(["league", "season", "game", "team"])
            .sort_index()
        )
        return df

    def read_lineup(  # noqa: C901
        self, match_id: Optional[Union[int, List[int]]] = None
    ) -> pd.DataFrame:
        """Retrieve lineups for the selected leagues and seasons.

        Parameters
        ----------
        match_id : int or list of int, optional
            Retrieve the lineup for a specific game.

        Raises
        ------
        ValueError
            If no games with the given IDs were found for the selected seasons and leagues.

        Returns
        -------
        pd.DataFrame.
        """
        urlmask = ESPN_API + "/{}/summary?event={}"
        filemask = "Summary_{}.json"

        df_schedule = self.read_schedule().reset_index()
        if match_id is not None:
            iterator = df_schedule[
                df_schedule.game_id.isin([match_id] if isinstance(match_id, int) else match_id)
            ]
            if len(iterator) == 0:
                raise ValueError(
                    "No games with the given IDs found for the selected seasons and leagues."
                )
        else:
            iterator = df_schedule

        df_list = []
        for i, match in iterator.iterrows():
            url = urlmask.format(match["league_id"], match["game_id"])
            filepath = self.data_dir / filemask.format(match["game_id"])
            reader = self.get(url, filepath)

            data = json.load(reader)
            for i in range(2):
                if "roster" not in data["rosters"][i]:
                    logger.info(
                        "No lineup info found for team %d in game with ID=%s",
                        i + 1,
                        match["game_id"],
                    )
                    continue
                for p in data["rosters"][i]["roster"]:
                    match_sheet = {
                        "game": match["game"],
                        "league": match["league"],
                        "season": match["season"],
                        "team": data["boxscore"]["form"][i]["team"]["displayName"],
                        "is_home": (i == 0),
                        "player": p["athlete"]["displayName"],
                        "position": p["position"]["name"] if "position" in p else None,
                        "formation_place": p["formationPlace"] if "formationPlace" in p else None,
                    }

                    if p["starter"]:
                        match_sheet["sub_in"] = "start"
                    elif p["subbedIn"]:
                        ii = [i for i, x in enumerate(p["plays"]) if x["substitution"]][0]
                        match_sheet["sub_in"] = sum(
                            map(
                                int,
                                re.findall(
                                    r"(\d{1,3})",
                                    p["plays"][ii]["clock"]["displayValue"],
                                ),
                            )
                        )
                    else:
                        match_sheet["sub_in"] = None

                    if (p["starter"] or p["subbedIn"]) and not p["subbedOut"]:
                        match_sheet["sub_out"] = "end"
                    elif p["subbedOut"]:
                        j = 0 if not p["subbedIn"] else 1
                        ii = [i for i, x in enumerate(p["plays"]) if x["substitution"]][j]
                        match_sheet["sub_out"] = sum(
                            map(
                                int,
                                re.findall(
                                    r"(\d{1,3})",
                                    p["plays"][ii]["clock"]["displayValue"],
                                ),
                            )
                        )
                    else:
                        match_sheet["sub_out"] = None

                    if "stats" in p:
                        for stat in p["stats"]:
                            match_sheet[stat["name"]] = stat["value"]

                    df_list.append(match_sheet)

        if len(df_list) == 0:
            return pd.DataFrame()

        df = (
            pd.DataFrame(df_list)
            .replace({"team": TEAMNAME_REPLACEMENTS})
            .pipe(standardize_colnames)
            .set_index(["league", "season", "game", "team", "player"])
            .sort_index()
        )
        return df



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['soccerdata']

package_data = \
{'': ['*']}

install_requires = \
['PySocks>=1.7.1,<2.0.0',
 'Unidecode>=1.2.0,<2.0.0',
 'flake8-docstrings>=1.6.0,<2.0.0',
 'flake8-rst-docstrings>=0.3.0,<0.4.0',
 'html5lib>=1.1,<2.0',
 'lxml>=4.6,<5.0',
 'pandas>=1.0,<2.0',
 'pretty-errors>=1.2.25,<2.0.0',
 'requests>=2.23,<3.0',
 'rich>=13.0.0,<14.0.0',
 'selenium>=4.0.0,<5.0.0',
 'undetected-chromedriver>=3.1.3,<4.0.0',
 'unicode>=2.7,<3.0']

setup_kwargs = {
    'name': 'soccerdata',
    'version': '1.3.2',
    'description': 'A collection of wrappers over soccer data from various websites / APIs.',
    'long_description': ".. image:: https://raw.githubusercontent.com/probberechts/soccerdata/master/docs/_static/logo2.png\n   :align: center\n   :alt: SoccerData\n   :width: 600px\n\n.. badges-begin\n\n|PyPI| |Python Version| |License| |Read the Docs| |Tests| |Codecov| |pre-commit| |Black|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/soccerdata.svg\n   :target: https://pypi.org/project/soccerdata/\n   :alt: PyPI\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/soccerdata\n   :target: https://pypi.org/project/soccerdata\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/pypi/l/soccerdata.svg\n   :target: https://opensource.org/licenses/Apache-2.0\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/soccerdata/latest.svg?label=Read%20the%20Docs\n   :target: https://soccerdata.readthedocs.io/\n   :alt: Read the documentation at https://soccerdata.readthedocs.io/\n.. |Tests| image:: https://github.com/probberechts/soccerdata/workflows/CI/badge.svg\n   :target: https://github.com/probberechts/soccerdata/actions?workflow=CI\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/probberechts/soccerdata/branch/master/graph/badge.svg\n   :target: https://app.codecov.io/gh/probberechts/soccerdata\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n\n.. badges-end\n\nSoccerData is a collection of wrappers over soccer data from `Club Elo`_,\n`ESPN`_, `FBref`_, `FiveThirtyEight`_, `Football-Data.co.uk`_, `SoFIFA`_ and\n`WhoScored`_. You get Pandas DataFrames with sensible, matching column names\nand identifiers across datasets. Data is downloaded when needed and cached\nlocally.\n\n.. code:: python\n\n   import soccerdata as sd\n\n   # Create scraper class instance for the Premier League\n   five38 = sd.FiveThirtyEight('ENG-Premier League', '1819')\n\n   # Fetch dataframes\n   games = five38.read_games()\n\nTo learn how to install, configure and use SoccerData, see the\n`Quickstart guide <https://soccerdata.readthedocs.io/en/latest/usage.html>`__. For documentation on each of the\nsupported data sources, see the `example notebooks <https://soccerdata.readthedocs.io/en/latest/datasources/>`__ and `API reference <https://soccerdata.readthedocs.io/en/latest/reference/>`__.\n\n.. _Club Elo: https://www.clubelo.com/\n.. _ESPN: https://www.espn.com/soccer/\n.. _FBref: https://www.fbref.com/en/\n.. _FiveThirtyEight: https://fivethirtyeight.com/soccer-predictions/\n.. _Football-Data.co.uk: https://www.football-data.co.uk/\n.. _SoFIFA: https://sofifa.com/\n.. _WhoScored: https://www.whoscored.com/\n\n**Disclaimer:** As soccerdata relies on web scraping, any changes to the\nscraped websites will break the package. Hence, do not expect that all code\nwill work all the time. If you spot any bugs, then please `fork it and start\na pull request <https://github.com/probberechts/soccerdata/blob/master/CONTRIBUTING.rst>`__.\n",
    'author': 'Pieter Robberechts',
    'author_email': 'pieter.robberechts@kuleuven.be',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/probberechts/soccerdata',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.2,<4.0.0',
}


setup(**setup_kwargs)
